from importlib import resources as impresources
import numpy as np
from pydrake.all import (
    AddMultibodyPlantSceneGraph,
    ConstantVectorSource,
    Frame,
    InverseDynamicsController,
    ModelInstanceIndex,
    MultibodyPlant,
    Multiplexer,
    Parser,
    RigidTransform,
)
from pydrake.geometry import SceneGraph, Meshcat, MeshcatVisualizer
from pydrake.math import RollPitchYaw, RigidTransform, RotationMatrix
from pydrake.multibody.tree import FixedOffsetFrame
from pydrake.systems.framework import Diagram, DiagramBuilder
from pydrake.systems.primitives import Demultiplexer
from typing import Tuple

# Local imports
from brom_drake.robots.gripper_type import GripperType
from brom_drake.control.grippers.gripper_controller import GripperController
from brom_drake.file_manipulation.urdf import DrakeReadyURDFConverter
from brom_drake.utils.leaf_systems import EndEffectorWrenchCalculator

from brom_drake import robots

class UR10eStation(Diagram):
    """
    *Description*

    A template system diagram for controlling a UR10e robot in a simulated environment.
    
    *Diagram*
    
    The diagram's inputs and outputs can be represented as: ::

                                            -----------------
                                            |               |
        gripper_target -------------------->|               |
        (Vector, np.array)                  |   UR10e       |
                                            |               |
        gripper_target_type --------------->|   Station     |
        (Abstract, GripperTarget Enum)      |               |
                                            |               |
        desired_joint_positions ----------->|               |
        (Vector, np.array)                  |               |
                                            |               |
                                            -----------------
    """

    def __init__(
        self,
        time_step: float = 0.002,
        meshcat_port_number: int = 7001,
        gripper_type: GripperType = GripperType.Robotiq_2f_85,
    ):
        """
        *Description*

        This class defines the UR10e Station.
        """
        # Input Processing
        self.meshcat_port_number = meshcat_port_number
        self.time_step = time_step

        # Initialize the Diagram
        Diagram.__init__(self)
        self.set_name("UR10e_Station")

        # Create a diagram builder
        self.builder = DiagramBuilder()

        # Create the plant and scene graph
        self.plant, self.scene_graph = None, None
        self.controller_plant = None
        self.create_plants_and_scene_graph(time_step)

        # Body ID's and Poses for Anything else in the scene
        self.object_ids = []
        self.object_poses = []

        # Whether we have a camera in the simulation
        # self.has_camera = False
        self.AddArm()

        # Which sort of gripper we're using (if any)
        self.gripper_type = gripper_type
        self.gripper_controller = None
        if gripper_type == GripperType.Robotiq_2f_85:
            self.Add2f85Gripper()
            self.add_gripper_controller()


    def add_arm_to_plant_with_ee_frame(
        self,
        arm_urdf_path: str,
        plant: MultibodyPlant,
        new_frame_name: str = "end_effector",
        X_ee: RigidTransform = RigidTransform(),
    ) -> Tuple[ModelInstanceIndex, Frame]:
        """
        *Description*
        
        This function adds the "arm" object at the arm_urdf_path location to the plant.

        *Parameters*
        
        arm_urdf_path : str
            The path to the URDF file for the arm.
        plant : MultibodyPlant
            The plant to which the arm will be added.
        
        *Returns*
        
        arm: ModelInstanceIndex
            The model instance index of the arm and the frame of the end effector.
        
        end_effector_frame: Frame
            The frame of the end effector.
        """
        # Setup

        # Add the arm to the provided plant
        arm = Parser(plant=plant).AddModels(arm_urdf_path)[0]

        # Fix the base of the arm to the world
        plant.WeldFrames(
            plant.world_frame(),
            plant.GetFrameByName("base_link", arm),
        )

        # Add Frame for the end effector
        end_effector_frame = plant.AddFrame(
            FixedOffsetFrame(
                new_frame_name,
                plant.GetFrameByName(self.end_effector_frame_name, arm),
                X_ee,
                arm,
            )
        )

        return arm, end_effector_frame
    
    def add_gripper_controller(self):
        """
        *Description*
        
        This function adds a controller for the gripper to the station.
        It is optional and may not be necessary for all stations.
        """
        # Setup

        # Create gripper controller
        self.gripper_controller = self.builder.AddSystem(
            GripperController(self.gripper_type)
        )
        self.gripper_controller.set_name(f"{self.get_name()}_gripper_controller")

    def AddArm(self):
        """
        *Description*
        
        Add the UR10e arm to the system.
        """
        # Setup
        arm_urdf_path = str(
            impresources.files(robots) / "models/ur/ur10e.urdf",
        )
        arm_urdf = DrakeReadyURDFConverter(arm_urdf_path).convert_urdf()
        arm_urdf = str(arm_urdf)

        self.end_effector_frame_name = "tool0"

        # Create transform for EE
        self.X_ee = RigidTransform()
        self.X_ee.set_translation([0, 0, 0.13])

        # Add the arm to the plant using our convenience function
        self.arm, self.arm_ee_frame = self.add_arm_to_plant_with_ee_frame(
            arm_urdf, self.plant,
            X_ee=self.X_ee,
        )

        # Add the arm to the controller plant using our convenience function
        self.controller_arm, self.controller_arm_ee_frame = self.add_arm_to_plant_with_ee_frame(
            arm_urdf, self.controller_plant,
            X_ee=self.X_ee,
        )

    def Add2f85Gripper(self):
        """
        *Description*

        Add the Robotiq 2F-85 gripper to the system. The arm must be added first.
        """
        # Setup

        # Add a gripper with actuation to the full simulated plant
        gripper_urdf_path = str(
            impresources.files(robots) / "models/robotiq/2f_85_gripper/urdf/robotiq_2f_85.urdf"
        )
        self.gripper = Parser(plant=self.plant).AddModels(gripper_urdf_path)[0]

        X_grip = RigidTransform()
        X_grip.set_rotation(
            RotationMatrix(RollPitchYaw([0.0, 0.0, np.pi / 2]))
        )
        self.plant.WeldFrames(
            self.plant.GetFrameByName("tool0", self.arm),
            self.plant.GetFrameByName("robotiq_arg2f_base_link", self.gripper),
            X_grip,
        )

        # # Add a gripper without actuation to the controller plant
        # gripper_static_urdf = str(
        #     impresources.files(robots) / "models/robotiq/2f_85_gripper-no-mimic/urdf/robotiq_2f_85_static.urdf"
        # )
        # static_gripper = Parser(plant=self.controller_plant).AddModels(
        #     gripper_static_urdf
        # )[0]

        # self.controller_plant.WeldFrames(
        #     self.controller_plant.GetFrameByName("tool0", self.controller_arm),
        #     self.controller_plant.GetFrameByName("robotiq_arg2f_base_link", static_gripper),
        #     X_grip)
    
    def ConnectToMeshcatVisualizer(self, port=None):
        self.meshcat = Meshcat(port)
        m = MeshcatVisualizer(self.meshcat)
        m.AddToBuilder(self.builder, self.scene_graph, self.meshcat)

        print("Open %s in a browser to view the meshcat visualizer." % self.meshcat.web_url())

    def CreateArmControllerAndConnect(self) -> InverseDynamicsController:
        """
        Description
        -----------
        This function creates a Cartesian arm controller and connects it to the rest of the system.
        :return:
        """
        # Setup
        kp = 1.0e3 * np.ones((self.plant.num_actuated_dofs(self.arm),))
        kd = 2.0 * np.sqrt(kp)

        # Create InverseDynamicsController
        self.controller = self.builder.AddSystem(
            InverseDynamicsController(
                self.controller_plant,
                kp=kp,
                ki=0.0*kp,
                kd=kd,
                has_reference_acceleration=False,
            )
        )

        # Torques from controller go to the "true" plant
        self.builder.Connect(
            self.controller.get_output_port_control(),
            self.plant.get_actuation_input_port(self.arm),
        )
        self.builder.Connect(
            self.controller.get_output_port_control(),
            self.controller_plant.get_actuation_input_port(self.controller_arm),
        )

        # Connect plant's state to controller
        self.builder.Connect(
            self.plant.get_state_output_port(self.arm),
            self.controller.get_input_port_estimated_state(),
        )

        return self.controller

    def CreateGripperControllerAndConnect(self):
        """
        *Description*
        
        Connects the gripper controller to:
        - The Gripper Actuators (done via the plant of the station)
        
        and exports some of the gripper controller's inputs and outputs
        (so that they become the inputs and outputs of the station).
        """
        # Setup
        gripper_controller = self.gripper_controller

        # Export the inputs of the gripper controller to the diagram
        self.builder.ExportInput(
            gripper_controller.GetInputPort("gripper_target"),
            "gripper_target",
        )
        self.builder.ExportInput(
            gripper_controller.GetInputPort("gripper_target_type"),
            "gripper_target_type",
        )

        # Connect gripper controller to the plant
        self.builder.Connect(
            self.plant.get_state_output_port(self.gripper),
            gripper_controller.GetInputPort("gripper_state"),
        )
        self.builder.Connect(
            gripper_controller.GetOutputPort("applied_gripper_torque"),
            self.plant.get_actuation_input_port(self.gripper),
        )

        # Send gripper position and velocity as an output
        # self.builder.ExportOutput(
        #     gripper_controller.GetOutputPort("measured_gripper_position"),
        #     "measured_gripper_position",
        # )
        # self.builder.ExportOutput(
        #     gripper_controller.GetOutputPort("measured_gripper_velocity"),
        #     "measured_gripper_velocity",
        # )

    def create_plants_and_scene_graph(self, time_step: float = 0.001):
        """
        *Description*
        
        This function creates the plant and scene graph for the diagram.

        *Parameters*
        
        time_step : float
            The time step for the MultibodyPlant.
        """
        # Setup

        # Algorithm
        self.plant, self.scene_graph = AddMultibodyPlantSceneGraph(
            builder=self.builder,
            time_step=time_step,
        )
        self.plant.set_name(f"{self.get_name()}_Plant")
        self.scene_graph.set_name(f"{self.get_name()}_SceneGraph")

        # Create plant for controller
        self.controller_plant = self.builder.AddSystem(
            MultibodyPlant(time_step=time_step)
        )
        self.controller_plant.set_name(f"{self.get_name()}_ControllerPlant")


    def ExportArmPositionInputPort(self):
        """
        *Description*
        
        Creates the input port for the arm position (i.e. the joint position)
        for the station.
        This requires creating a special connection to the arm controller
        via a multiplexer.
        """
        # Setup
        controller = self.controller
        controller_plant = controller.get_multibody_plant_for_control()

        # Create multiplexer to provide input to arm controller
        mux_for_controller = self.builder.AddSystem(
            Multiplexer([
                self.plant.num_positions(self.arm),
                self.plant.num_velocities(self.arm),
            ]),
        )

        # Create constant velocity source
        constant_velocity_source = self.builder.AddSystem(
            ConstantVectorSource(
                np.zeros((controller_plant.num_actuated_dofs(self.arm),)),
            ),
        )

        # Connect the multiplexer to the controller
        self.builder.Connect(
            constant_velocity_source.get_output_port(),
            mux_for_controller.get_input_port(1),
        )
        self.builder.Connect(
            mux_for_controller.get_output_port(),
            controller.get_input_port_desired_state(),
        )

        # Export the second input port of the multiplexer
        self.builder.ExportInput(
            mux_for_controller.get_input_port(0),
            "desired_joint_positions",
        )

    def ExportArmControllerPorts(self):
        """
        *Description*
        
        This function creates the input and output ports for the diagram
        that are related to the arm controller.
        """
        # Setup
        controller = self.controller
        controller_plant = controller.get_multibody_plant_for_control()

        # Create commanded arm torque output
        self.builder.ExportOutput(
            self.controller.get_output_port_control(),
            "ur10e_arm_torque_commanded",
        )

        # Create desired position input
        self.ExportArmPositionInputPort()
        
        # Create Multiplexer to extract the arm position and velocity
        demux = self.builder.AddSystem(
            Demultiplexer(
                self.plant.num_multibody_states(self.arm),
                self.plant.num_positions(self.arm),
            ),
        )
        demux.set_name("UR10e_Station_Demux")

        self.builder.Connect(
            self.plant.get_state_output_port(self.arm),
            demux.get_input_port(0),
        )
        self.builder.ExportOutput(
            demux.get_output_port(0),
            "ur10e_arm_position_measured",
        )
        self.builder.ExportOutput(
            demux.get_output_port(1),
            "ur10e_arm_velocity_measured",
        )

        # Compute and output end-effector wrenches based on measured joint torques
        wrench_calculator = self.builder.AddSystem(
            EndEffectorWrenchCalculator(
                controller_plant,
                controller_plant.GetFrameByName("end_effector"),
            ),
        )
        wrench_calculator.set_name("wrench_calculator")

        self.builder.Connect(
            demux.get_output_port(0),
            wrench_calculator.GetInputPort("joint_positions"))
        self.builder.Connect(
            demux.get_output_port(1),
            wrench_calculator.GetInputPort("joint_velocities"))
        self.builder.Connect(
            self.controller.get_output_port_control(),
            wrench_calculator.GetInputPort("joint_torques"))

        self.builder.ExportOutput(
            wrench_calculator.get_output_port(),
            "measured_ee_wrench",
        )

    def ExportGripperStatePort(self):
        """
        *Description*
        
        This function exports the gripper's state
        as a port of the diagram.
        """
        # Setup

        # Export the state of the gripper
        self.builder.ExportOutput(
            self.plant.get_state_output_port(self.gripper),
            "gripper_state",
        )
    
    def Finalize(self):
        """
        *Description*
        
        This finalizes the diagram by:

        - Finalizing all plants in the diagram
        - Setting up the visualizer (if desired)
        - Creating and connecting the arm controller
        - Creating and connecting the gripper controller (if desired)
        - Building the diagram
        """
        # Setup

        # Finalize all plants
        self.plant.Finalize()
        self.controller_plant.Finalize()

        # # Set up the scene graph
        # self.builder.Connect(
        #     self.scene_graph.get_query_output_port(),
        #     self.plant.get_geometry_query_input_port()
        # )

        # self.builder.Connect(
        #     self.plant.get_geometry_pose_output_port(),
        #     self.scene_graph.get_source_pose_port(self.plant.get_source_id())
        # )

        # Connect Meshcat, if desired
        if self.use_meshcat():
            self.ConnectToMeshcatVisualizer(port=self.meshcat_port_number)

        # Create Arm and Gripper Controllers
        self.CreateArmControllerAndConnect()
        self.ExportArmControllerPorts()

        if self.gripper_type != GripperType.NoGripper:
            self.CreateGripperControllerAndConnect()
            self.ExportGripperStatePort()

        # Build the diagram
        self.builder.BuildInto(self)
    
    def use_meshcat(self) -> bool:
        """
        *Description*
        
        This function returns whether or not to use meshcat for visualization.
        :return: (bool) Whether or not to use meshcat for visualization.
        """
        return self.meshcat_port_number is not None


from importlib import resources as impresources
from pathlib import Path

import numpy as np
from pydrake.all import (
    Context,
    Frame,
    ModelInstanceIndex,
)
from pydrake.geometry import SceneGraph, Meshcat, MeshcatVisualizer, MeshcatVisualizerParams
from pydrake.geometry import Role as DrakeRole
from pydrake.math import RollPitchYaw, RigidTransform, RotationMatrix
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import MultibodyPlant, AddMultibodyPlantSceneGraph
from pydrake.multibody.tree import FixedOffsetFrame
from pydrake.systems.framework import Diagram, DiagramBuilder
from pydrake.systems.primitives import Demultiplexer
from typing import Tuple

# Local imports
from brom_drake.robots.gripper_type import GripperType
from brom_drake.control import IdealJointPositionController
from brom_drake.control.grippers.gripper_controller import GripperController
from brom_drake.file_manipulation.urdf import (
    DrakeReadyURDFConverter, 
    DrakeReadyURDFConverterConfig,
    MeshReplacementStrategy,
    MeshReplacementStrategies,
)

from brom_drake import robots

class UR10eStation(Diagram):
    """
    *Description*

    A template system diagram for controlling a UR10e robot in a simulated environment.

    *Diagram*
    
    """
    def __init__(
        self,
        time_step: float = 0.002,
        gripper_type: GripperType = GripperType.NoGripper,
        force_conversion_of_original_urdf: bool = False,
        meshcat_port_number: int = None,
        end_effector_frame_name: str = "tool0",
    ):
        """
        *Description*

        This class defines the KINEMATIC UR10e Manipulation Station.
        This station is based off of the Drake Manipulation Station showcased
        in Russ Tedrake's manipulation course.
        """
        # Input Processing
        self.force_conversion_of_original_urdf = force_conversion_of_original_urdf
        self.meshcat_port_number = meshcat_port_number
        self.end_effector_frame_name = end_effector_frame_name

        # Initialize the Diagram
        Diagram.__init__(self)
        self.set_name("UR10e_Station")

        # Create a diagram builder
        self.builder = DiagramBuilder()

        # Create scene_graph and plant
        self.plant, self.scene_graph = None, None
        self.controller_plant = None
        self.time_step = time_step
        self.create_plants_and_scene_graph()

        # Body ID's and Poses for Anything else in the scene
        self.object_ids = []
        self.object_poses = []

        # Whether we have a camera in the simulation
        # self.has_camera = False
        self.arm = None # Assign the Arm's Model Index later
        self.AddArm()

        # Which sort of gripper we're using (if any)
        self.gripper_type = gripper_type
        self.gripper = None
        self.gripper_controller = None
        if gripper_type == GripperType.Robotiq_2f_85:
            self.Add2f85Gripper()
            self.add_gripper_controller()

        # Visualization
        self.meshcat = None

    def add_arm_to_plant_with_ee_frame(
        self,
        arm_urdf_path: str,
        plant: MultibodyPlant,
        new_frame_name: str = "end_effector",
        X_ee: RigidTransform = RigidTransform(),
    ) -> Tuple[ModelInstanceIndex, Frame]:
        """
        *Description*
        
        This function adds the "arm" object at the arm_urdf_path location to the plant.

        *Parameters*
        
        arm_urdf_path : str
            The path to the URDF file for the arm.

        plant : MultibodyPlant
            The plant to which the arm will be added.

        new_frame_name : str, optional
            The name of the new end-effector frame to be added, by default "end_effector".

        X_ee : RigidTransform, optional
            The transform from the arm's end-effector frame to the new end-effector frame,
            by default RigidTransform().

        *Returns*

        arm : ModelInstanceIndex
            The model instance index of the added arm.

        new_frame : Frame
            The newly added end-effector frame.
        """
        # Setup

        # Add the arm to the provided plant
        arm = Parser(plant=plant).AddModels(arm_urdf_path)[0]

        # Fix the base of the arm to the world
        plant.WeldFrames(
            plant.world_frame(),
            plant.GetFrameByName("base_link", arm),
        )

        # Add Frame for the end effector
        new_frame = plant.AddFrame(
            FixedOffsetFrame(
                new_frame_name,
                plant.GetFrameByName(self.end_effector_frame_name, arm),
                X_ee,
                arm,
            )
        )

        return arm, new_frame

    def Add2f85Gripper(self):
        """
        Add the Robotiq 2F-85 gripper to the system. The arm must be added first.
        """
        # Setup

        # Add a gripper with actuation to the full simulated plant
        gripper_urdf_path = str(
            impresources.files(robots) / "models/robotiq/2f_85_gripper/urdf/robotiq_2f_85.urdf"
        )
        self.gripper = Parser(plant=self.plant).AddModels(gripper_urdf_path)[0]

        X_grip = RigidTransform()
        X_grip.set_rotation(
            RotationMatrix(RollPitchYaw([0.0, 0.0, np.pi / 2]))
        )
        self.plant.WeldFrames(
            self.plant.GetFrameByName("tool0", self.arm),
            self.plant.GetFrameByName("robotiq_arg2f_base_link", self.gripper),
            X_grip,
        )

        # Add a gripper without actuation to the controller plant
        gripper_static_urdf = str(
            impresources.files(robots) / "models/robotiq/2f_85_gripper/urdf/robotiq_2f_85_static.urdf"
        )
        static_gripper = Parser(plant=self.controller_plant).AddModels(
            gripper_static_urdf
        )[0]

        self.controller_plant.WeldFrames(
            self.controller_plant.GetFrameByName("tool0", self.controller_arm),
            self.controller_plant.GetFrameByName("robotiq_arg2f_base_link", static_gripper),
            X_grip,
        )

    def add_gripper_controller(self):
        """
        Description
        -----------
        This funciton adds a controller for the gripper to the station.
        It is optional and may not be necessary for all stations.
        """
        # Setup

        # Create gripper controller
        self.gripper_controller = self.builder.AddSystem(
            GripperController(self.gripper_type)
        )
        self.gripper_controller.set_name(f"{self.get_name()}_gripper_controller")

    def AddArm(self):
        """
        Add the UR10e arm to the system.
        """
        # Setup
        original_arm_urdf_path = str(
            impresources.files(robots) / "models/ur/ur10e.urdf",
        )
        expected_arm_urdf_path = Path("./brom/models/ur10e/ur10e.drake.urdf")

        # Define Transform from tool0 to end_effector frame
        self.X_ee = RigidTransform()
        self.X_ee.set_translation([0, 0, 0.13])

        # Convert the arm urdf if necessary
        arm_urdf = None
        if (not expected_arm_urdf_path.exists()) or self.force_conversion_of_original_urdf:
            # If drake-compatible URDF does not exist, then we need to convert
            # the original URDF to a drake-compatible URDF.
            config = DrakeReadyURDFConverterConfig(
                overwrite_old_logs=True,
                mesh_replacement_strategies=MeshReplacementStrategies(
                    collision_meshes=MeshReplacementStrategy.kWithObj,
                )
            )
            arm_urdf = DrakeReadyURDFConverter(
                original_arm_urdf_path,
                config=config,
            ).convert_urdf()
            arm_urdf = str(arm_urdf)
        else:
            # Otherwise, let's just read the drake-compatible URDF
            arm_urdf = str(expected_arm_urdf_path)

        # Add the arm to the plant using our convenience function
        self.arm, self.arm_ee_frame = self.add_arm_to_plant_with_ee_frame(
            arm_urdf, self.plant,
            X_ee=self.X_ee,
        )

        # Add the arm to the controller plant using our convenience function
        self.controller_arm, self.controller_arm_ee_frame = self.add_arm_to_plant_with_ee_frame(
            arm_urdf, self.controller_plant,
            X_ee=self.X_ee,
        )

    def connect_gripper_controller(self):
        """
        *Description*
        
        Connects the gripper controller to:
        - The Gripper Actuators (done via the plant of the station)
        
        and exports some of the gripper controller's inputs and outputs
        (so that they become the inputs and outputs of the station).
        """
        # Setup
        gripper_controller = self.gripper_controller

        # Export the inputs of the gripper controller to the diagram
        self.builder.ExportInput(
            gripper_controller.GetInputPort("gripper_target"),
            "gripper_target",
        )
        self.builder.ExportInput(
            gripper_controller.GetInputPort("gripper_target_type"),
            "gripper_target_type")

        # Connect gripper controller to the plant
        self.builder.Connect(
            self.plant.get_state_output_port(self.gripper),
            gripper_controller.GetInputPort("gripper_state"),
        )
        self.builder.Connect(
            gripper_controller.GetOutputPort("applied_gripper_torque"),
            self.plant.get_actuation_input_port(self.gripper),
        )

        # Send gripper position and velocity as an output
        self.builder.ExportOutput(
            gripper_controller.GetOutputPort("measured_gripper_position"),
            "measured_gripper_position")
        self.builder.ExportOutput(
            gripper_controller.GetOutputPort("measured_gripper_velocity"),
            "measured_gripper_velocity",
        )

    def ConnectToMeshcatVisualizer(self, port: int=None):
        """
        *Description*

        This function connects the station to a Meshcat visualizer.

        *Parameters*

        port: int, optional
            The port number for the Meshcat server.
            If None, the default port will be used.
        """
        self.meshcat = Meshcat(port)
        m = MeshcatVisualizer(self.meshcat)
        m.AddToBuilder(
            self.builder, self.scene_graph, self.meshcat,
            # params=MeshcatVisualizerParams(
            #         role=DrakeRole.kProximity,
            #     ),
        )

        print("Open %s in a browser to view the meshcat visualizer." % self.meshcat.web_url())

    def create_plants_and_scene_graph(self):
        """
        *Description*
        
        This function creates the plant for the UR10e station, if needed.
        """
        # Setup

        # Create SceneGraph
        # self.scene_graph = self.builder.AddSystem(
        #     SceneGraph()
        # )
        # self.scene_graph.set_name(f"{self.get_name()}_SceneGraph")

        # # Create plant (will contain ALL elements of scene)
        # self.plant = self.builder.AddSystem(
        #     MultibodyPlant(time_step=self.time_step)
        # )
        # self.plant.RegisterAsSourceForSceneGraph(self.scene_graph)
        # self.plant.set_name(f"{self.get_name()}_Plant")

        self.plant, self.scene_graph = AddMultibodyPlantSceneGraph(
            builder=self.builder,
            time_step=self.time_step,
        )
        self.plant.set_name(f"{self.get_name()}_Plant")
        self.scene_graph.set_name(f"{self.get_name()}_SceneGraph")

        # Create plant for controller
        self.controller_plant = MultibodyPlant(time_step=self.time_step)
        self.controller_plant.set_name(f"{self.get_name()}_ControllerPlant")

    def Finalize(self):
        """
        *Description*
        
        This finalizes the diagram by:
        - Finalizing all plants in the diagram
        - Setting up the scene graph connections
        - Connecting to Meshcat (if desired)
        - Creating the arm and gripper controllers
        - Building the diagram
        """
        # Setup

        # Announce that we are finalizing the station
        # print("Finalizing station!")

        # Finalize all plants
        self.plant.Finalize()
        self.controller_plant.Finalize()

        # Set up the scene graph
        # self.builder.Connect(
        #     self.scene_graph.get_query_output_port(),
        #     self.plant.get_geometry_query_input_port()
        # )

        # self.builder.Connect(
        #     self.plant.get_geometry_pose_output_port(),
        #     self.scene_graph.get_source_pose_port(self.plant.get_source_id())
        # )

        self.builder.ExportOutput(
            self.scene_graph.get_query_output_port(),
            "query_object",
        )

        # Connect Meshcat, if desired
        if self.use_meshcat():
            self.ConnectToMeshcatVisualizer(port=self.meshcat_port_number)

        # Create Arm and Gripper Controllers
        self.CreateArmPorts()

        if self.gripper_type != GripperType.NoGripper:
            self.connect_gripper_controller()

        # Build the diagram
        self.builder.BuildInto(self)

    def CreateArmPorts(
        self,
    ):
        """
        *Description*
        
        This function creates a Cartesian arm controller and connects it to the rest of the system.
        :return:
        """
        # Setup

        # Create the input ports for the joint positions
        self.arm_controller = self.CreateJointArmControllerAndExportIO()

        # Output measured arm position and velocity
        demux = self.builder.AddSystem(
            Demultiplexer(self.plant.num_multibody_states(self.arm), self.plant.num_positions(self.arm)),
        )
        demux.set_name(f"{self.get_name()}_Demux")

        self.builder.Connect(
            self.plant.get_state_output_port(self.arm),
            demux.get_input_port(0),
        )
        self.builder.ExportOutput(
            demux.get_output_port(0),
            "measured_arm_position",
        )
        self.builder.ExportOutput(
            demux.get_output_port(1),
            "measured_arm_velocity",
        )

    def CreateJointArmControllerAndExportIO(
        self,
    )-> IdealJointPositionController:
        """
        *Description*
        
        Creates an ideal joint position controller and exports the necessary inputs + outputs.
        
        *Returns*

        joint_controller: IdealJointPositionController
            The created joint position controller.
        """
        # Setup
        arm = self.arm
        plant = self.plant

        # Create the joint controller
        joint_controller = self.builder.AddSystem(
            IdealJointPositionController(arm, plant),
        )

        # Export the input and output ports
        self.builder.ExportInput(
            joint_controller.desired_joint_positions_port,
            "desired_joint_positions",
        )

        self.builder.ExportOutput(
            joint_controller.GetOutputPort("measured_joint_positions"),
            "measured_joint_positions",
        )

        return joint_controller

    def use_meshcat(self) -> bool:
        """
        *Description*

        This function returns whether or not Meshcat visualization is being used.

        *Returns*

        meshcat_should_be_used: bool
            True if Meshcat visualization is being used, False otherwise.
        """
        return self.meshcat_port_number is not None
    
    def UpdateInternalContexts(self, context: Context):
        """
        *Description*

        This function updates the internal contexts of the plant and the controller.

        TODO(Kwesi): Clean up this method of unused variables.

        *Parameters*

        context: pydrake.systems.framework.Context
            The context from which to update the internal contexts.
        """
        # Setup
        plant_context = self.plant.GetMyMutableContextFromRoot(context)
        controller_context = self.controller_plant.GetMyMutableContextFromRoot(context)

        # Update the contexts
        self.arm_controller.plant_context = plant_context
        # self.controller_plant.SetDefaultState(controller_context)


from dataclasses import dataclass
from enum import IntEnum
import numpy as np
from typing import List, Union

class FSMTransitionConditionType(IntEnum):
    """
    Enum to represent the different types of conditions that can be used to
    transition between states in a finite state machine.
    """
    kEqual = 0
    kGreaterThanEqual = 1
    kLessThanEqual = 2
    kAfterThisManySeconds = 3

class FSMTransitionCondition:
    """
    Dataclass to represent a single condition that can be used to transition
    between states in a finite state machine.
    """
    def __init__(
        self,
        condition_type: FSMTransitionConditionType,
        condition_value: Union[bool, float, np.ndarray],
        input_port_name: Union[str, None] = None,
    ):
        """
        Description
        -----------
        Constructor for the FSMTransitionCondition class.
        """
        # Setup

        # Input Checking
        assert condition_type in FSMTransitionConditionType, "Invalid condition type provided."

        # Assign Attributes
        self.input_port_name = input_port_name
        self.type = condition_type
        self.condition_value = condition_value

        # Check that the condition contains the input_port_name IF the condition type
        # is not kAfterThisManySeconds
        if self.type != FSMTransitionConditionType.kAfterThisManySeconds:
            assert self.input_port_name is not None, \
                f"Input port name must be provided for condition type \"{self.type}\"."
            
        # Check that the condition value is a positive float IF the condition type is
        # kAfterThisManySeconds
        if self.type == FSMTransitionConditionType.kAfterThisManySeconds:
            assert isinstance(self.condition_value, float), \
                f"Condition value must be a float for condition type \"{self.type}\"."
            assert self.condition_value > 0.0, \
                f"Condition value must be a positive float for condition type \"{self.type}\"."

    @staticmethod
    def ForAfterThisManySeconds(seconds: float) -> "FSMTransitionConditionType":
        """Convenience method for creating transition conditions that are \'after `seconds` seconds\'"""
        # Setup
        return FSMTransitionCondition(
            condition_type=FSMTransitionConditionType.kAfterThisManySeconds,
            condition_value=seconds,
        )

    def evaluate_comparison(self, input_port_value: Union[bool, float, np.ndarray]):
        """
        Description
        -----------
        This function evaluates the condition based on the input port value.
        """
        # Setup
        input_x = input_port_value
        if (type(input_x) == int) or (type(input_x) == float):
            input_x = np.ndarray([input_x])


        # Input Checking
        if self.type == FSMTransitionConditionType.kAfterThisManySeconds:
            raise NotImplementedError(
                f"Condition type \"kAfterThisManySeconds\" can not yet be evaluated yet implemented."
            )

        # Evaluate the condition for array types
        if type(input_x) == np.ndarray:
            if self.type == FSMTransitionConditionType.kEqual:
                return np.all(input_x == self.condition_value)
            elif self.type == FSMTransitionConditionType.kGreaterThanEqual:
                return np.all(input_x >= self.condition_value)
            elif self.type == FSMTransitionConditionType.kLessThanEqual:
                return np.all(input_x <= self.condition_value)
            else:
                raise NotImplementedError("Condition type not yet implemented.")
        else:
            # Fallback for all other types of comparisons
            if self.type == FSMTransitionConditionType.kEqual:
                return input_port_value == self.condition_value
            elif self.type == FSMTransitionConditionType.kGreaterThanEqual:
                return input_port_value >= self.condition_value
            elif self.type == FSMTransitionConditionType.kLessThanEqual:
                return input_port_value <= self.condition_value
            else:
                raise NotImplementedError("Condition type not yet implemented.")
            
    def requires_input_port(self):
        """
        Description
        -----------
        This function returns whether or not the condition requires an input port.
        """
        return self.type != FSMTransitionConditionType.kAfterThisManySeconds


from typing import List, Tuple

import numpy as np
from pathlib import Path
from pydrake.geometry import Meshcat, MeshcatVisualizer, MeshcatVisualizerParams
from pydrake.geometry import Role as DrakeRole
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import AddMultibodyPlantSceneGraph, MultibodyPlant
from pydrake.systems.framework import DiagramBuilder, Diagram, Context

from pydrake.systems.primitives import ConstantVectorSource, VectorLogSink

# Internal Imports
from brom_drake.robots import find_base_link_name_in
from brom_drake.productions.types import BaseProduction
from brom_drake.productions import ProductionID
from brom_drake.productions.roles.role import Role
from brom_drake.utils import Performer, AddMultibodyTriad
from .show_me_system import ShowMeSystem

class ShowMeThisModel(BaseProduction):
    """
    **Description**
    
    A production which allows the user to visualize a specified model
    in Meshcat, with the ability to set desired joint positions, if the model is
    a robot/articulated.
    
    **Parameters**
    
    path_to_model: str
        The path to the model file to be visualized.
        TODO: Add support for Path type. Make sure to add tests!!

    with_these_joint_positions: List[float], optional
        A list of joint positions to set the model to when visualized.
        The length of this list must match the number of positions in the model.
        If None, the model will be visualized in its default configuration.
        Default is None.

    base_link_name: str, optional
        The name of the base link of the model. If None, the production will
        attempt to automatically determine the base link name.
        Default is None.

    time_step: float, optional
        The time step to use for the MultibodyPlant.
        Default is 1e-3.

    meshcat_port_number: int, optional
        The port number to use for Meshcat visualization.
        If None, Meshcat visualization will be disabled.
        Default is 7001.

    show_collision_geometries: bool, optional
        A boolean flag indicating whether to show collision geometries in Meshcat
        INSTEAD of the visual geometries.
        Default is False.

    **Usage**

    The production can be used to show both simple models (e.g., single objects)
    or articulated models (e.g., robots). 
    
    To show a simple model, simply provide the path to the model file. ::

        from brom_drake.productions import ShowMeThisModel

        # Some stuff here

        urdf_file_path = "path/to/your/model.urdf"

        # Create the production
        production = ShowMeThisModel(
            urdf_file_path,
            time_step=1e-3,
        )

        # Add the cast to the production, build the diagram, and simulate
        diagram, diagram_context = production.add_cast_and_build()

        # Set up simulation
        simulator = Simulator(diagram, diagram_context)
        simulator.set_target_realtime_rate(1.0)
        simulator.set_publish_every_time_step(False)

    If the model is articulated, then you can also provide desired joint positions. ::

        from brom_drake.productions import ShowMeThisModel

        # Some stuff here

        urdf_file_path = "path/to/your/articulated_model.urdf"
        desired_joint_positions = [0.0, 1.0, -1.0, 0.5]  # Example joint positions

        # Create the production
        production = ShowMeThisModel(
            urdf_file_path,
            with_these_joint_positions=desired_joint_positions,
            time_step=1e-3,
        )

        # Add the cast to the production, build the diagram, and simulate
        diagram, diagram_context = production.add_cast_and_build()

        # Set up simulation
        simulator = Simulator(diagram, diagram_context)
        simulator.set_target_realtime_rate(1.0)
        simulator.set_publish_every_time_step(False)

    """
    def __init__(
        self,
        path_to_model: str|Path,
        with_these_joint_positions: List[float] = None,
        base_link_name: str = None,
        time_step: float = 1e-3,
        meshcat_port_number: int = 7001, # Usually turn off for CI (i.e., make it None)
        show_collision_geometries: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)

        # Copy all parameters into the production
        if type(path_to_model) is str:
            path_to_model = Path(path_to_model)

        self.path_to_model = path_to_model
        self.q_des = with_these_joint_positions
        self.base_link_name = base_link_name
        self.time_step = time_step
        self.meshcat_port_number = meshcat_port_number
        self.show_collision_geometries = show_collision_geometries

        # Check that the model exists
        if not self.path_to_model.exists():
            raise ValueError(f"The path \"{self.path_to_model}\" does not exist! Can not show a model that does not exist!")

        # If the base link name is not provided,
        # then we will try to do a smart search for it.
        if self.base_link_name is None:
            self.base_link_name = find_base_link_name_in(self.path_to_model)

        # Add Plant and Scene Graph for easy simulation
        self.plant, self.scene_graph = AddMultibodyPlantSceneGraph(
            self.builder,
            time_step=self.time_step,
        )
        self.plant.set_name("ShowMeThisModel_plant")

        self.meshcat = None
        self.model_index, self.model_name = None, None
        self.show_me_system = None

    def add_supporting_cast(self):
        """
        **Description**
        
        This method updates the production's
        - MultibodyPlant with the user's model
        - Using the ShowMeSystem to freeze the model in place
        - Connects to Meshcat, if requested

        This is required to be implemented by all children of the BaseProduction class.
        """
        # Setup

        # Add Model
        model_idcs = Parser(plant=self.plant).AddModels(
            str(self.path_to_model)
        )
        assert len(model_idcs) == 1, \
            f"Only one model should be added in this production;" + \
            f" received {len(model_idcs)} in the file {self.path_to_model}."
        self.model_index = model_idcs[0]
        self.model_name = self.plant.GetModelInstanceName(model_idcs[0])

        # Collect the expected number of actuated joints
        n_dofs = self.find_number_of_positions_in_model()
        if self.q_des is None:
            self.q_des = np.zeros(n_dofs)
        else:
            assert len(self.q_des) == n_dofs, \
                f"Expected {n_dofs} joint positions; received {len(self.q_des)}."

        # Add ShowMeSystem
        self.show_me_system = self.builder.AddSystem(
            ShowMeSystem(
                plant=self.plant,
                model_index=self.model_index,
                desired_joint_positions=np.array(self.q_des),
            ),
        )

        self.create_and_connect_source_of_joint_positions()

        # Add Sink
        output_joints_sink = self.builder.AddSystem(
            VectorLogSink(len(self.q_des)),
        )

        # Connect the system to the sink
        self.builder.Connect(
            self.show_me_system.get_output_port(0),
            output_joints_sink.get_input_port(0),
        )

        self.add_multibody_triads()

        # Weld the base link to the world frame
        self.plant.WeldFrames(
            self.plant.world_frame(),
            self.plant.GetFrameByName(self.base_link_name, self.model_index),
        )

        # Connect to Meshcat, if requested
        if self.meshcat_port_number is not None:
            self.meshcat = Meshcat(port=self.meshcat_port_number)  # Object provides an interface to Meshcat
            m_visualizer = MeshcatVisualizer(
                self.meshcat,
            )
            params = MeshcatVisualizerParams(
                role=DrakeRole.kIllustration,
            )
            if self.show_collision_geometries:
                params = MeshcatVisualizerParams(
                    role=DrakeRole.kProximity,
                )

            m_visualizer.AddToBuilder(
                self.builder, self.scene_graph, self.meshcat,
                params=params,
            )

        # Finalize plant and connect it to system
        self.plant.Finalize()

    def add_cast_and_build(
        self,
        cast: Tuple[Role, Performer] = [],
    ) -> Tuple[Diagram, Context]:
        """
        **Description**
        
        This method builds the production's diagram and context,
        and assigns the plant context to the internal ShowMeSystem.
        
        **Parameters**
        
        cast: Tuple[Role, Performer], optional
            The cast to be added to the production.
            Default is an empty tuple.
            
        **Returns**

        diagram: Diagram
            The built Diagram of the production.

        diagram_context: Context
            The built Context of the production.
        """
        super().add_cast_and_build(cast)

        # Assign the diagram context to the internal show_me_system
        self.show_me_system.mutable_plant_context = self.plant.GetMyMutableContextFromRoot(
            self.diagram_context,
        )

        return self.diagram, self.diagram_context

    def add_multibody_triads(self):
        """
        **Description**
        
        This method will add triads to the world frame
        and any other relevant frames.
        """
        # Setup

        # Input Checking
        assert self.model_index is not None, \
            "The model index is not yet set. Please call add_supporting_cast() first."

        # Add A Triad to base?
        AddMultibodyTriad(self.plant.world_frame(), self.scene_graph)

    def create_and_connect_source_of_joint_positions(self):
        """
        **Description**
        
        This method will create and connect a source of joint positions
        to the ShowMeSystem.

        **Assumptions**

        - ``self.show_me_system`` is not None
        - ``self.q_des`` is not None
        """
        # Setup

        assert self.show_me_system is not None, \
            "The show_me_system is not yet created. Please call add_supporting_cast() first."

        assert self.q_des is not None, \
            "The desired joint positions are not yet set. Please provide them at initialization."

        # Add Source
        desired_joint_positions_source = self.builder.AddSystem(
            ConstantVectorSource(np.array(self.q_des)),
        )

        # Connect the source to the system
        self.builder.Connect(
            desired_joint_positions_source.get_output_port(0),
            self.show_me_system.get_input_port(0),
        )

    def find_number_of_positions_in_model(
        self,
    ) -> int:
        """
        **Description**
        
        This method will return the number of positions in the model
        as defined by the user (through the path_to_model).

        **Returns**
        
        n_positions: int
            The number of positions in the model.
        """
        # Setup

        # Create a shadow plant
        shadow_plant = MultibodyPlant(self.time_step)
        model_idcs = Parser(plant=shadow_plant).AddModels(
            str(self.path_to_model)
        )
        shadow_model_idx = model_idcs[0]

        # Weld the base link to the world frame
        shadow_plant.WeldFrames(
            shadow_plant.world_frame(),
            shadow_plant.GetFrameByName(self.base_link_name, shadow_model_idx),
        )

        # Finalize the shadow plant
        shadow_plant.Finalize()

        return shadow_plant.num_positions()

    @property
    def id(self) -> ProductionID:
        """
        **Description**
        
        This property returns the unique identifier for this production.
        (i.e., ProductionID.kShowMeThisModel)

        This is required to be implemented by all children of the BaseProduction class.

        **Returns**
        
        id: ProductionID
            The unique identifier for this production.
        """
        return ProductionID.kShowMeThisModel

    @property
    def suggested_roles(self) -> List[Role]:
        """
        **Description**

        This property returns an empty list, because there are no suggested
        roles for this production.

        This is required to be implemented by all children of the BaseProduction class.

        **Returns**

        roles: List[Role]
            A list of suggested roles for this production.
        """
        return []



from dataclasses import dataclass
import networkx as nx
import numpy as np
from pydrake.all import RigidTransform, PiecewisePose
from typing import List

# Internal Imports
from brom_drake.utils.leaf_systems.network_fsm import (
    NetworkXFSM, FSMOutputDefinition, FSMTransitionCondition, FSMTransitionConditionType
)
from .phases import AttemptGraspWithPuppeteerWristPhase

# Define some internal classes/enums
@dataclass
class Script:
    """
    *Description*
    
    This class defines the "script" or the sequence of events that is meant to happen in the
    AttemptGrasp production.

    *Attributes*

    settling_time_on_floor: float, optional
        The time (in seconds) to let the object settle on the floor before the gripper approaches.
        Default is 10.0 seconds.

    gripper_approach_time: float, optional
        The time (in seconds) for the gripper to approach the object.
        Default is 5.0 seconds.

    grasp_closing_time: float, optional
        The time (in seconds) for the gripper to close around the object.
        Default is 2.0 seconds.

    post_grasp_settling_time: float, optional
        The time (in seconds) to let the object settle in the grasp after the gripper has closed.
        Default is 0.1 seconds.

    drop_time: float, optional
        The time (in seconds) to let the object fall after releasing the grasp.
        Default is 10.0 seconds.
    """
    settling_time_on_floor: float = 10.0
    gripper_approach_time: float = 5.0
    grasp_closing_time: float = 2.0
    post_grasp_settling_time: float = 0.1
    drop_time: float = 10.0

    def add_all_states_to_networkx_graph(self, graph: nx.DiGraph) -> None:
        """
        *Description*
        
        This method adds all the states to the provided NetworkX graph.
        
        .. note::
            :collapsible:

            One of the states (kObjectSettlingInGrasp) is not explicitly created
            here, but it will be created anyway when the edges are added. (This is a feature
            of NetworkX where nodes can be created implicitly when edges are added.)

        *Parameters*
        
        graph: nx.DiGraph
            A blank NetworkX directed graph to which the states will be added.

        """
        # Setup

        # First State: Object settled
        graph.add_node(
            AttemptGraspWithPuppeteerWristPhase.kObjectSettlingOnFloor,
            outputs=[
                FSMOutputDefinition("start_floor", False),  # Floor trigger
                FSMOutputDefinition("enable_gripper_approach", True),  # Make gripper move towards object
                FSMOutputDefinition("close_gripper", False),  # Gripper trigger
                FSMOutputDefinition("switch_to_memory", "trajectory1"),  # Switch to memory system for gripper pose
            ]
        )

        # Second State: Gripper approach
        graph.add_node(
            AttemptGraspWithPuppeteerWristPhase.kGripperApproach,
            outputs=[
                FSMOutputDefinition("enable_gripper_approach", True),  # Make gripper move towards object
                FSMOutputDefinition("start_floor", False),  # Floor trigger
                FSMOutputDefinition("close_gripper", False),  # Gripper trigger
            ]
        )

        # Third State: Close gripper
        graph.add_node(
            AttemptGraspWithPuppeteerWristPhase.kGripperClosing,
            outputs=[
                FSMOutputDefinition("enable_gripper_approach", True),  # Stop gripper movement
                FSMOutputDefinition("close_gripper", True),
            ]
        )

        # Fourth State: Drop the floor
        graph.add_node(
            AttemptGraspWithPuppeteerWristPhase.kFloorDrop,
            outputs=[
                FSMOutputDefinition("start_floor", True),
                FSMOutputDefinition("switch_to_memory", "memory2"),  # Switch to memory system for gripper pose
            ]
        )

    def build_gripper_approach_trajectory_from_waypoints(self, waypoints_X_WorldGripper: List[RigidTransform]) -> PiecewisePose:
        """
        *Description*
        
        Returns a PiecewisePose trajectory for the gripper base to follow during the
        attempted grasp.

        The trajectory should be an interpolation between two poses that is then held
        until the end of the production.

        *Parameters*

        waypoints_X_WorldGripper: List[RigidTransform]
            The waypoints that the script will create a Pose trajectory for.
            This trajectory will start at the initial waypoint and 

        *Returns*

        gripper_base_trajectory: PiecewisePose
            A PiecewisePose trajectory for the gripper base to follow during the
            attempted grasp.
        """

        # Setup
        n_waypoints = len(waypoints_X_WorldGripper)

        # Compute the Grasp Pose of the gripper in the world frame
        pose_WorldGripper_initial = waypoints_X_WorldGripper[0]

        # For each event in the script, create (i) a time and (ii) a pose
        times = []
        poses = []
        
        # - Create initial pose
        times.append(0.0)
        poses.append(pose_WorldGripper_initial)

        # - Force all poses between (a) pre-grasp pose (will be same as initial pose) and (b) the final grasp pose
        #   to occur during the time window between kGripperApproach and kGripperClosing
        t_pre_grasp = self.start_time_of_phase(
            phase=AttemptGraspWithPuppeteerWristPhase.kGripperApproach
        )
        t_gripper_closing = self.start_time_of_phase(
            phase=AttemptGraspWithPuppeteerWristPhase.kGripperClosing
        )

        times_during_interpolation = np.linspace(start=t_pre_grasp, stop=t_gripper_closing, num=n_waypoints)
        for idx, time_i in enumerate(times_during_interpolation):
            times.append(time_i)
            poses.append(waypoints_X_WorldGripper[idx])

        # Assemble into a PiecewisePose
        gripper_base_trajectory = PiecewisePose.MakeLinear(
            times=times,
            poses=poses,
        )
        return gripper_base_trajectory

    def start_time_of_phase(self, phase: AttemptGraspWithPuppeteerWristPhase) -> float:
        """
        *Description*
        
        This method returns the start time of the given phase.
        
        *Parameters*
        
        phase: AttemptGraspWithPuppeteerWristPhase
            The phase for which to get the start time.
            
        *Returns*
        
        start_time: float
            The start time of the given phase in seconds.
        """
        match phase:
            case AttemptGraspWithPuppeteerWristPhase.kObjectSettlingOnFloor:
                return 0.0
            case AttemptGraspWithPuppeteerWristPhase.kGripperApproach:
                return self.settling_time_on_floor
            case AttemptGraspWithPuppeteerWristPhase.kGripperClosing:
                return self.settling_time_on_floor + self.gripper_approach_time
            case AttemptGraspWithPuppeteerWristPhase.kObjectSettlingInGrasp:
                return self.settling_time_on_floor + self.gripper_approach_time + self.grasp_closing_time
            case AttemptGraspWithPuppeteerWristPhase.kFloorDrop:
                return self.settling_time_on_floor + self.gripper_approach_time + self.grasp_closing_time + self.post_grasp_settling_time
            case _:
                raise ValueError(f"Unrecognized phase {phase}!")

    def to_fsm(self) -> NetworkXFSM:
        """
        *Description*

        This method converts the script to a NetworkXFSM instance.

        *Returns*

        fsm: NetworkXFSM
            The NetworkXFSM LeafSystem which coordinates/sends triggers out according to the script.
        """
        return NetworkXFSM(self.to_networkx_graph())

    def to_networkx_graph(self) -> nx.DiGraph:
        """
        *Description*
        
        This method converts the script to a NetworkX directed graph.

        *Returns*
        
        graph: nx.DiGraph
            The NetworkX directed graph representing the FSM of the script.
        """
        # Setup
        graph = nx.DiGraph()

        # Create NetworkX FSM to trigger the floor trajectory
        self.add_all_states_to_networkx_graph(graph)

        # Create time from init -> Let object settle
        graph.add_edge(
            AttemptGraspWithPuppeteerWristPhase.kObjectSettlingOnFloor,
            AttemptGraspWithPuppeteerWristPhase.kGripperApproach, 
            conditions=[
                FSMTransitionCondition(
                    condition_type=FSMTransitionConditionType.kAfterThisManySeconds,
                    condition_value=self.settling_time_on_floor,
                )
            ]
        )

        # Create time between gripper approach and start og grasp closing
        graph.add_edge(
            AttemptGraspWithPuppeteerWristPhase.kGripperApproach,
            AttemptGraspWithPuppeteerWristPhase.kGripperClosing, 
            conditions=[
                FSMTransitionCondition(
                    condition_type=FSMTransitionConditionType.kAfterThisManySeconds,
                    condition_value=self.gripper_approach_time,
                )
            ]
        )

        if self.post_grasp_settling_time > 0.0:
            # Create time from gripper start to end of grasp closing
            graph.add_edge(
                AttemptGraspWithPuppeteerWristPhase.kGripperClosing,
                AttemptGraspWithPuppeteerWristPhase.kObjectSettlingInGrasp,
                conditions=[
                    FSMTransitionCondition(
                        condition_type=FSMTransitionConditionType.kAfterThisManySeconds,
                        condition_value=self.grasp_closing_time,
                    )
                ]
            )

            # Create time from end of grasp closing to floor drop start
            graph.add_edge(
                AttemptGraspWithPuppeteerWristPhase.kObjectSettlingInGrasp,
                AttemptGraspWithPuppeteerWristPhase.kFloorDrop,
                conditions=[
                    FSMTransitionCondition(
                        condition_type=FSMTransitionConditionType.kAfterThisManySeconds,
                        condition_value=self.post_grasp_settling_time,
                    )
                ]
            )

        else:
            # Create time from end of grasp closing to end of production
            graph.add_edge(
                AttemptGraspWithPuppeteerWristPhase.kGripperClosing,
                AttemptGraspWithPuppeteerWristPhase.kFloorDrop,
                conditions=[
                    FSMTransitionCondition(
                        condition_type=FSMTransitionConditionType.kAfterThisManySeconds,
                        condition_value=self.grasp_closing_time,
                    )
                ]
            )

        return graph

    def total_time(self) -> float:
        """
        *Description*

        This method returns the total time of the script.

        *Returns*

        total_time: float
            The total time of the script in seconds.
        """
        return self.settling_time_on_floor + self.gripper_approach_time + self.grasp_closing_time + self.post_grasp_settling_time + self.drop_time
    



from typing import Union, Tuple, List

from pydrake.systems.framework import DiagramBuilder, LeafSystem, Diagram, Context

# Internal Imports
from brom_drake.PortWatcher.plotter import FigureNamingConvention
from brom_drake.productions.roles import Role
from brom_drake.productions.ids import ProductionID
from brom_drake.utils import Performer
from brom_drake.utils.watcher import add_watcher_and_build
from brom_drake.utils.initial_condition_manager import InitialConditionManager

class BaseProduction:
    """
    *Description*

    Base class for all productions.

    *Attributes*

    builder : DiagramBuilder
        The DiagramBuilder used to build the production.

    diagram : Diagram
        The Diagram of the production.
        This is None and must be built using the build_production method.

    diagram_context : Context
        The Context of the production's Diagram.
        This is None and must be built using the build_production method.

    performers : List[Performer]
        A list of all performers added to the production.

    watcher : Optional[DiagramWatcher]
        The DiagramWatcher for the production.
        This is None, unless the production is built with a watcher.

    plant : Optional[MultibodyPlant]
        The MultibodyPlant of the production, if applicable.
        This is None by default.

    scene_graph : Optional[SceneGraph]
        The SceneGraph of the production, if applicable.
        This is None by default.

    initial_condition_manager : InitialConditionManager
        The InitialConditionManager used to manage initial conditions
        for the production.
        
    """
    def __init__(self, **kwargs):
        # Create a builder for the production
        self.builder = DiagramBuilder()

        # Create the diagram and diagram_context placeholders
        self.diagram = None
        self.diagram_context = None

        # Save a list of the performers
        self.performers = []

        # Create an extra place to save DiagramWatcher objects
        self.watcher = None

        # Create a place to save the plant and scene_graph, if needed
        self.plant, self.scene_graph = None, None

        # Create an initial condition manager
        self.initial_condition_manager = InitialConditionManager()

    def add_supporting_cast(self):
        """
        *Description*
        
        This method must be implemented by the subclass.
        It should add all "secondary" elements to the builder (i.e., the
        cast members that the user doesn't need to worry about).
        
        """
        pass

    def suggested_roles(self) -> List[Role]:
        """
        *Description*
        
        This method should be implemented by the subclass.
        It should return a list of roles that are suggested for the production.
        
        *Returns*

        suggested_roles: List[Role]
            A list of roles that are suggested for the production.
        """
        return []

    def fill_role(
        self,
        role: Role,
        system: Performer,
    ):
        """
        *Description*
        
        This method should be implemented by the subclass. It should add the
        system to the role.
        
        *Parameters*

        role: Role
            The role to be filled.

        system: Performer
            The system to fill the role with.
        """
        # Setup
        builder = self.builder

        # Call the member method of the role object
        role.connect_performer_ports_to(builder, system)

        # Save the performer
        self.performers.append(system)

    def add_main_cast(
        self,
        cast: List[Tuple[Role, Performer]] = [],
    ):
        # Setup

        # Fulfill each role-performer pair in the casting_call list
        for role, performer in cast:
            self.fill_role(role, performer)

    def build_production(
        self,
        with_watcher: bool = True,
        figure_naming_convention: FigureNamingConvention = FigureNamingConvention.kFlat,
    ) -> Tuple[Diagram, Context]:
        """
        *Description*
        
        This method builds the production.
        It assumes that all components have been added to the builder.
        
        *Parameters*

        with_watcher: bool, optional
            Whether to build the production with a DiagramWatcher.
            Default is True.

        figure_naming_convention: FigureNamingConvention, optional
            The naming convention to use for figures in the watcher.
            Default is FigureNamingConvention.kFlat.

        *Returns*

        diagram: Diagram
            The built Diagram of the production.

        diagram_context: Context
            The Context of the built Diagram.
            If with_watcher is True, this will be the Context of the DiagramWatcher, as well.
        """
        # Setup
        builder = self.builder

        # Set all initial conditions, if any
        if self.plant is not None:
            self.initial_condition_manager.set_all_initial_conditions(
                plant=self.plant,
                # diagram_context=self.diagram_context,
            )

        # Build the diagram
        if with_watcher:
            self.watcher, self.diagram, self.diagram_context = add_watcher_and_build(
                builder,
                figure_naming_convention=figure_naming_convention
            )
        else:
            self.diagram = builder.Build()
            self.diagram_context = self.diagram.CreateDefaultContext()

        return self.diagram, self.diagram_context

    def add_cast_and_build(
        self,
        main_cast_members: Tuple[Role, Performer] = [],
        with_watcher: bool = True,
        figure_naming_convention: FigureNamingConvention = FigureNamingConvention.kFlat,
    ) -> Tuple[Diagram, Context]:
        # Setup

        # Add the cast to the production
        self.add_supporting_cast()
        self.add_main_cast(cast=main_cast_members)

        return self.build_production(
            with_watcher=with_watcher,
            figure_naming_convention=figure_naming_convention,
        )

    @property
    def id(self) -> ProductionID:
        """
        *Description*
        
        The unique identifier for the production.

        If not defined, returns ProductionID.kNotDefined.
        
        *Returns*
        
        ProductionID
            The unique identifier for the production.
        """
        return ProductionID.kNotDefined


"""
Description:
    This file defines a Role class that is used to populate the
    production with ONLY systems that match the appropriate structure/signature.
"""
from dataclasses import dataclass
from typing import List

from pydrake.systems.framework import DiagramBuilder

from brom_drake.productions.roles.role_port_assignment import RolePortAssignment
# Internal Imports
from brom_drake.utils import Performer

@dataclass
class Role:
    """
    *Description*

    This class defines a "slot" to be filled in a Drake Diagram.
    A name is used for unique identification and a description is (optionally)
    used to describe what is expected of a drake LeafSystem (or Diagram) that
    properly "fills the role" (i.e., fills the slot) in a Drake Diagram.

    *Parameters*

    name: str
        Unique identifier of the role.

    description: str
        Non-binding description of the role.

    port_assignments: List[RolePortAssigment]
        Each element of this list describes:
        
        - an external target (e.g., `LeafSystems`) that the role expects to connect to
        - the port on the ``LeafSystem`` that fills this role that should connect to that target.

    *Usage*

    Consider the following pieces of code: ::

        from brom_drake.productions.roles import Role, RolePortAssignment
        from brom_drake.productions.roles.role_port_assignment import PairingType

        kGraspConfirmer = Role(
            name="Confirm-When-Grasp-Worked",
            description="Attempts to define with a boolean whether or not the object we're interested in is currently in our grasp.",
            port_assignments=[
                # Required inputs for Role
                RolePortAssignment(
                    external_target_name="target_cube_pose",
                    performer_port_name="current_manipuland_pose",
                    pairing_type=PairingType.kInput
                ),
                RolePortAssignment(
                    external_target_name="cube_model_idx",
                    performer_port_name="object_model_index",
                    pairing_type=PairingType.kInput
                ),
                RolePortAssignment(
                    external_target_name="gripper_pose",
                    performer_port_name="gripper_pose",
                    pairing_type=PairingType.kInput,
                ),
                RolePortAssignment(
                    external_target_name="production_gripper_model_idx",
                    performer_port_name="gripper_model_index",
                    pairing_type=PairingType.kInput
                ),
                RolePortAssignment(
                    external_target_name="gripper_configuration",
                    performer_port_name="gripper_configuration",
                    pairing_type=PairingType.kInput,
                ),
                # Output Of Role
                RolePortAssignment(
                    external_target_name="grasp_complete",
                    performer_port_name="grasp_success_flag",
                    pairing_type=PairingType.kOutput,
                )
            ]
        )

    
    The above example defines a Role with:

    - name: ``"Confirm-When-Grasp-Worked"``
    - Required Input Ports:
        + ``"current_manipuland_pose"``
        + ``"object_model_index"``
        + ``"gripper_pose"``
        + ``"gripper_model_index"``
        + ``"gripper_configuration"``
    - Required Output Port:
        + ``"grasp_success_flag"``
    - Required Ports to exist *somewhere else in the Diagram*:
        + ``"target_cube_pose"``
        + ``"cube_model_idx"``
        + ``"gripper_pose"``
        + ``"production_gripper_model_idx"``
        + ``"gripper_configuration"``
        + ``"grasp_complete"``

    Without all of these satisfied, we can not successfully place
    a `LeafSystem` into a Production that expects the ``"Confirm-When-Grasp-Worked"``
    role.

    """
    name: str
    description: str
    port_assignments: List[RolePortAssignment]

    def __str__(self):
        return self.name

    def connect_performer_ports_to(
        self,
        builder: DiagramBuilder,
        performer: Performer,
    ):
        """
        *Description*

        Connect each inputs and output port mentioned in `self.port_assignments`
        on `performer` to the first port found in `builder` with the correct name.

        .. warning ::

            The ``.Build()`` method should not have been called on ``builder``
            before invoking this method.
        
        *Parameters*

        builder: DiagramBuilder
            The ``DiagramBuilder`` containing a partially built diagram.

        performer: Performer
            The object that we think can fill the role defined by self.
        """
        # Setup

        # Add performer to builder
        builder.AddSystem(performer)

        # Connect performer to the rest of the production (stored in Builder)
        for assignment in self.port_assignments:
            assignment.create_connections_in(builder, performer)


import numpy as np
from pydrake.all import (
    AbstractValue, BasicVector, Context,
    LeafSystem,
    PiecewisePolynomial, PiecewiseQuaternionSlerp, PiecewisePose,
    RigidTransform,
)

# Internal Imports
from brom_drake.motion_planning.systems.proximity_pose_plan_dispenser import (
    ProximityPosePlanDispenserConfig,
    DispenserTransitionRequest,
)
from brom_drake.motion_planning.systems.proximity_pose_plan_dispenser.dispenser_internal_state import (
    DispenserInternalState, DispenserTransitionRequest,
)


class ProximityPosePlanDispenser(LeafSystem):
    """
    *Description*
    
    A dispenser which will dispense the plan that is given to it. It will provide the
    next point in the plan when the current_pose is "within proximity" of the current target
    in the plan. The proximity is defined by the proximity config that is given to the
    dispenser.

    *Diagram*

    The LeafSystem's inputs and outputs can be represented as: ::

                                    -----------------
                                    |               |
        current_pose -------------> |  Proximity    | ----> pose_in_plan
                                    |  Pose         |
        plan           -----------> |  Plan         | ----> dispenser_state
                                    |  Dispenser    |
        request ------------------> |               | ----> current_pose_index
                                    |               | 
                                    -----------------

    *Input Ports*
    
    plan: AbstractValue (List[RigidTransform])
        This port is used to receive the plan that the system should follow.
    plan_ready: AbstractValue (bool)
        This port is used to receive a boolean that indicates if the plan is ready.
        Lower this (to false) and then raise it (to true) to trigger the system to
        reset the plan.
    """
    def __init__(
        self,
        config: ProximityPosePlanDispenserConfig = ProximityPosePlanDispenserConfig(),
    ):
        """
        *Description*
        
        This function initializes the Proximity Pose Plan Dispenser.

        *Parameters*
        
        config: ProximityPosePlanDispenserConfig
            The configuration for the Proximity Pose Plan Dispenser.
            See brom_drake/motion_planning/systems/proximity_pose_plan_dispenser/proximity_pose_plan_dispenser.py
            for more information.
        """
        LeafSystem.__init__(self)
        # Setup
        self.config = config

        # Create empty variables for the plan and the planned trajectory
        self.plan = []

        # Create input and output ports for the plan
        self.plan_is_set = False
        self.t0, self.t_final = -1.0, -1.0
        self.plan_port, self.request_port = None, None
        self.declare_input_ports()
        self.declare_output_ports()

        # Create internal state
        self.dispenser_state = self.DeclareDiscreteState(
            np.array([DispenserInternalState.kReady])
        )
        self.dispenser_plan_index = 0

        self.last_pose = RigidTransform()

    def advance_plan_if_necessary(self, context: Context):
        """
        *Description*
        
        This function is responsible for advancing the plan index if the system
        is ready to do so.

        *Parameters*
        
        context: Context
            The current context of the system.
        """
        # Setup
        plan = self.plan_port.Eval(context)
        plan_idx = self.dispenser_plan_index
        current_pose = self.current_pose_port.Eval(context)

        state = context.get_discrete_state(self.dispenser_state)[0]

        # If the plan is not ready, then do nothing
        if state != DispenserInternalState.kPlanSet:
            return

        # If the plan is ready, then check to see if we should advance the plan
        should_advance = self.config.in_proximity(
            current_pose,
            plan[plan_idx],
        )

        if should_advance and (plan_idx < len(plan)-1):
            self.dispenser_plan_index += 1
                    

    def declare_input_ports(self):
        """
        *Description*

        This function creates the input ports for the plan dispenser.
        """
        # Setup

        # Define ports
        sample_plan = [RigidTransform()]
        self.current_pose_port = self.DeclareAbstractInputPort(
            "current_pose",
            AbstractValue.Make(RigidTransform()),
        )
        
        self.plan_port = self.DeclareAbstractInputPort(
            "plan",
            AbstractValue.Make(sample_plan),
        )

        self.request_port = self.DeclareVectorInputPort(
            "request",
            1,
        )

    def declare_output_ports(self):
        """
        *Description*
        
        This function creates the output ports for the plan dispenser.
        """
        # Setup

        # Define ports
        self.DeclareAbstractOutputPort(
            "pose_in_plan",
            lambda: AbstractValue.Make(RigidTransform()),
            self.GetCurrentPoseInPlan,
        )

        self.DeclareVectorOutputPort(
            "dispenser_state",
            size=1,
            calc=self.GetInternalState,
        )

        self.DeclareVectorOutputPort(
            "current_pose_index",
            size=1,
            calc=self.GetCurrentPoseIndex,
        )

    def GetCurrentPoseIndex(self, context: Context, output: BasicVector):
        """
        *Description*
        
        This function is responsible for outputting the current index of the plan that the system is on.

        *Parameters*
        
        context: Context
            The current context of the system.

        output: BasicVector
            The output of the system. We will assign a single value to this vector.
        """
        # Setup
        plan_idx = self.dispenser_plan_index

        # Output the current index
        output.SetFromVector(
            np.array([plan_idx])
        )

    def GetCurrentPoseInPlan(self, context: Context, output_pose: AbstractValue):
        """
        *Description*
        
        This function is responsible for outputting the current pose in the plan that the system is on.
        The plan will only advance if the system is currently "within proximity"
        of the current target in the plan. (i.e., if the current pose satisfies the
        in_proximity() function of the proximity config.)
        
        *Parameters*
        
        context: Context
            The current context of the system.

        output_pose: AbstractValue
            The output of the system. We will assign a single RigidTransform to this value.
        """
        # Setup
        plan = self.plan_port.Eval(context)
        last_pose = self.last_pose

        # Transition the internal state, if needed
        self.transition_internal_state(context)

        # Get the abstract state
        state = context.get_discrete_state(self.dispenser_state)

        # If the state of the dispenser is reset, then share
        # the LAST pose.
        dispenser_in_reset = state[0] == DispenserInternalState.kReady
        if dispenser_in_reset:
            output_pose.SetFrom(
                AbstractValue.Make(last_pose)
            )
            return
        
        # If the state of the plan is kPlanSet,
        # then check to see if we should ADVANCE to the next part of the plan
        self.advance_plan_if_necessary(context)
        plan_idx = self.dispenser_plan_index

        # Output The Current Point in the Plan
        output_pose.SetFrom(
            AbstractValue.Make(plan[plan_idx])
        )

    def GetInternalState(self, context: Context, output: BasicVector):
        """
        *Description*
        
        Retrieves the current value of the internal state of this
        planner. The internal state is defined by the DispenserInternalState
        enum found in brom_drake/motion_planning/systems/proximity_pose_plan_dispenser/dispsenser_internal_state.py.

        *Parameters*
        
        context: Context
            The current context of the system.

        output: BasicVector
            The output of the system. We will assign a single value to this vector.
        """
        # Setup

        # Set the internal state (if necessary)
        self.transition_internal_state(context)

        # Output the internal state
        output.SetFrom(
            context.get_discrete_state(self.dispenser_state)
        )

    def transition_internal_state(self, context: Context):
        """
        *Description*
        
        This function is responsible for transitioning the internal state of the system
        based on the request that is given to it.

        *Parameters*
        
        context: Context
            The current context of the system.
        """
        # Setup
        current_state = context.get_discrete_state(self.dispenser_state)[0]

        # Collect the request from the port
        request = self.request_port.Eval(context)

        # Transition the state
        dispenser_is_ready = current_state == DispenserInternalState.kReady
        dispenser_plan_is_set = current_state == DispenserInternalState.kPlanSet

        if (request == DispenserTransitionRequest.kRequestSavePlan) and dispenser_is_ready:
            # Save the plan AND transition to the "Plan Set" state
            current_plan = self.plan_port.Eval(context)
            self.plan = current_plan
            self.dispenser_plan_index = 0

            # Set the internal state
            context.SetDiscreteState(
                np.array([DispenserInternalState.kPlanSet])
            )
        elif (request == DispenserTransitionRequest.kRequestReset) and dispenser_plan_is_set:
            # Prepare the system to receive a new plan
            # self.dispenser_plan_index = 0

            # Set the internal state
            context.SetDiscreteState(
                np.array([DispenserInternalState.kReady])
            )
        else:
            # Otherwise, do nothing 
            pass

    # def initialize_system_for_new_plan(self, context: Context):
    #     """
    #     Description
    #     -----------
    #     This function should lock a couple of things into memory for this object to handle future
    #     calls to get the plan.

    #     :param context:
    #     :return:
    #     """
    #     # Setup

    #     # Get time to start executing the plan
    #     self.t0 = context.get_time()
    #     context.SetDiscreteState(
    #         np.array([DispenserInternalState.kSet])
    #     )

    #     # Get times where we should reach each point in the plan and save it into a new map
    #     self.plan = self.plan_port.Eval(context)

    #     n_points_in_plan = len(self.plan)
    #     t_ii = 0.0
    #     times = [t_ii]
    #     positions_as_array = np.zeros((0, 3))
    #     for ii in range(n_points_in_plan-1):
    #         t_ii += self.find_time_between_two_points(
    #             self.plan[ii].translation(),
    #             self.plan[ii+1].translation(),
    #             )
            
    #         # Save the times, positions and orientaitons separately
    #         times += [t_ii]
    #         positions_as_array = np.vstack(
    #             (positions_as_array, self.plan[ii].translation()),
    #         )

    #     # Save final time, position and orientation
    #     self.t_final = t_ii
    #     positions_as_array = np.vstack(
    #         (positions_as_array, self.plan[-1].translation()),
    #     )

    #     # Create the planned trajectory
    #     position_trajectory = PiecewisePolynomial.FirstOrderHold(
    #         times,
    #         [p.translation() for p in self.plan],
    #     )
    #     orientation_trajectory = PiecewiseQuaternionSlerp(times, [RigidTransform(p).rotation() for p in self.plan])
        
    #     self.planned_trajectory = PiecewisePose(
    #         position_trajectory,
    #         orientation_trajectory,
    #     )





"""
This file defines a couple of convenience functions for adding in the watcher
to a target diagram builder. These functions should be useful for 99% of the
users of the DiagramWatcher class.
"""
from typing import List, Tuple, Union

import numpy as np
from pydrake.all import DiagramBuilder, Diagram
from pydrake.systems.framework import Context

from brom_drake.DiagramTarget import DiagramTarget
from brom_drake.DiagramWatcher import DiagramWatcher, DiagramWatcherOptions
from brom_drake import directories
from brom_drake.PortWatcher.port_watcher_options import (
    PortFigureArrangement,
    PortWatcherPlottingOptions,
    PortWatcherRawDataOptions,
    FigureNamingConvention,
    PortWatcherOptions,
)

PotentialTargetTypes = List[
    Union[
        str,
        Tuple[str, int],
        Tuple[int, int],
        Tuple[str, str],
        Tuple[str, List[int]],
        Tuple[int, List[int]],
        Tuple[str, List[str]],
    ],
]


def add_watcher(
    builder: DiagramBuilder,
    targets: PotentialTargetTypes = None,
    watcher_dir: str = directories.DEFAULT_WATCHER_DIR,
    plot_arrangement: PortFigureArrangement = PortFigureArrangement.OnePlotPerPort,
    figure_naming_convention: FigureNamingConvention = FigureNamingConvention.kFlat,
    file_format: str = "png",
) -> DiagramWatcher:
    """
    **Description**

    This function adds a :py:class:`~brom_drake.DiagramWatcher.DiagramWatcher.DiagramWatcher` to a Drake Diagram.
    The diagram is not finalized yet and so it is accessible via the `builder` argument.
    The watcher will insert LogVectorSink systems to log the output ports of the systems specified in the `targets` argument.

    Importantly, this function does NOT build the diagram.

    **Parameters**

    builder : DiagramBuilder
        The diagram builder to which we want to add the watcher.

    targets : List[Tuple[Union[str, int]]]
        The targets that we want to watch.

    watcher_dir : str, optional
        The directory in which we will store the data collected by the DiagramWatcher.
        By default, :py:const:`~brom_drake.directories.directories.DEFAULT_WATCHER_DIR`.

    plot_arrangement : PortFigureArrangement, optional
        The arrangement of the plots.
        By default, PortFigureArrangement.OnePlotPerPort.

    figure_naming_convention : FigureNamingConvention, optional
        The naming convention for the figures.
        By default, FigureNamingConvention.kFlat.

    file_format : str, optional
        The file format for the figures.
        By default, "png".

    **Example Usage**

    The simplest way to use this function is to just pass in the diagram builder:

    .. code-block:: python

        watcher = add_watcher(builder)

    You can also specify which systems and ports to watch by passing in a list of targets.
    Each target can be specified as either:

    - A string representing the name of the system to watch (all ports will be watched), or
    - A tuple of (system_name: str, port_index: int) to watch a specific port of a system, or
    - A tuple of (system_index: int, port_index: int) to watch a specific port of a system by its index.

    Here are some examples of how to specify targets:

    .. code-block:: python

        watcher = add_watcher(builder, [("plant", 0), ("controller", 0)])

    **Returns**
        
    watcher : DiagramWatcher
        The watcher that we have added to the diagram builder.
    """
    # Input Processing
    if not isinstance(plot_arrangement, PortFigureArrangement):
        raise ValueError(
            f"plot_arrangement must be of type PortFigureArrangement; received {plot_arrangement} of type" +
            f" {type(plot_arrangement)}."
        )

    # Parse targets list if it exists
    if targets is not None:
        targets = parse_list_of_simplified_targets(builder, targets)

    watcher = DiagramWatcher(
        builder,
        targets=targets,
        options=DiagramWatcherOptions(
            base_directory=watcher_dir,
            plotting_options=PortWatcherPlottingOptions(
                plot_arrangement=plot_arrangement,
                figure_naming_convention=figure_naming_convention,
                file_format=file_format,
            ),
            raw_data_options=PortWatcherRawDataOptions(
                save_to_file=True,
            ),
        )
    )
    return watcher


def add_watcher_and_build(
    builder: DiagramBuilder,
    targets: PotentialTargetTypes = None,
    watcher_dir: str = directories.DEFAULT_WATCHER_DIR,
    plot_arrangement: PortFigureArrangement = PortFigureArrangement.OnePlotPerPort,
    figure_naming_convention: FigureNamingConvention = FigureNamingConvention.kFlat,
    file_format: str = "png",
) -> Tuple[DiagramWatcher, Diagram, Context]:
    """
    **Description**

    This function adds a :py:class:`~brom_drake.DiagramWatcher.DiagramWatcher.DiagramWatcher` to a Drake Diagram.
    The diagram is not finalized yet and so it is accessible via the `builder` argument.
    The watcher will insert LogVectorSink systems to log the output ports of the systems specified in the `targets` argument.

    Once all of the LogVectorSink systems have been added, the diagram is built and a reference to the built diagram
    and its default context is added to the DiagramWatcher object.

    Example Usage:

    .. code-block:: python

        watcher = add_watcher_and_build(builder)

    In addition, one can specify which systems to watch:

    .. code-block:: python

        watcher = add_watcher_and_build(builder, targets=[("plant",)])

    The DiagramWatcher will then watch all supported ports of the "plant" system
    that it can find.

    Furthermore, to specify which systems and specific ports to watch:

    .. code-block:: python

        watcher = add_watcher_and_build(builder, targets=[("plant", 0), ("controller", 0)])

    This will watch output port 0 of both the "plant" and "controller" systems, if possible.

    **Parameters**
    
    builder: DiagramBuilder
        The diagram builder to which we want to add the watcher.

    targets: List[Tuple[Union[str, int]]]
        The targets that we want to watch.

    watcher_dir: str
        The directory in which we will store the data collected by the DiagramWatcher.

    plot_arrangement: PortFigureArrangement
        The arrangement of the plots.
        (Can be PortFigureArrangement.OnePlotPerPort OR PortFigureArrangement.OnePlotPerDim)

    figure_naming_convention: FigureNamingConvention
        The naming convention for the figures.
        (Can be FigureNamingConvention.kFlat OR FigureNamingConvention.kHierarchical)

    file_format: str
        The file format for the figures.
    
    **Returns**

    watcher: DiagramWatcher
        The watcher that we have added to the diagram builder.

    diagram: Diagram
        The built diagram.

    diagram_context: Context
        The default context for the built diagram.
    """
    watcher = add_watcher(
        builder,
        targets=targets,
        watcher_dir=watcher_dir,
        plot_arrangement=plot_arrangement,
        figure_naming_convention=figure_naming_convention,
        file_format=file_format,
    )

    # Build the diagram and add a reference to the watcher
    diagram = builder.Build()
    diagram_context = diagram.CreateDefaultContext()
    watcher.diagram = diagram
    watcher.diagram_context = diagram_context

    return watcher, diagram, diagram_context


def parse_list_of_simplified_targets(
    builder: DiagramBuilder,
    targets: PotentialTargetTypes,
) -> List[DiagramTarget]:
    """
    **Description**
    
    This function takes a list of simplified targets and converts them to the full form.

    **Example Usage**
    
    .. code-block:: python

        targets = [("plant", 0), ("controller", 0)]
        parsed_targets = parse_list_of_simplified_targets(targets)

    **Parameters**
    
    builder : DiagramBuilder
        The diagram builder to which we want to add the watcher.

    targets : List[Tuple[Union[str, int]]
        The list of simplified targets.

    **Returns**
    
    parsed_targets: List[DiagramTarget]
        A list of all the targets that we want to watch.
    """
    # Setup
    parsed_targets = []
    system_list = builder.GetSystems()

    # If input is not a list, then raise an error
    if not isinstance(targets, list):
        raise ValueError("the input \"targets\" must be a list.")

    if len(targets) == 0:
        raise ValueError(
            "the input \"targets\" must be a non-empty list containing tuples or strings.\n" +
            "If you want to watch the entire diagram, then pass in None as the value of targets."
        )

    # Parse each element in targets
    for target in targets:
        target_name, ports_list = None, None

        if isinstance(target, str):
            target_name = target
        elif isinstance(target, tuple):
            # Parse the first element in the tuple
            if isinstance(target[0], str):
                target_name = target[0]
            elif isinstance(target[0], int):
                # Check if the index is valid
                if target[0] < 0 or target[0] >= len(system_list):
                    raise ValueError(
                        f"the index {target[0]} is not a valid index for the list of systems (has length {len(system_list)})."
                    )
                target_name = system_list[target[0]].get_name()

            else:
                raise ValueError(
                    "the first element of the tuple must be a string or an integer; received {} (type {}).".format(
                        target[0], type(target[0])
                    )
                )

            # Parse the second element in the tuple
            if isinstance(target[1], int):
                ports_list = [target[1]]
            elif isinstance(target[1], str):
                # Find the index of port with this name
                port_name = target[1]

                # Find the system
                system = None
                for ii, system_ii in enumerate(system_list):
                    if system_ii.get_name() == target_name:
                        system = system_ii
                        break

                # for jj in range(system.num_output_ports()):
                #     print(system.get_output_port(jj).get_name())

                if system.HasOutputPort(port_name):
                    ports_list = [int(system.GetOutputPort(port_name).get_index())]
                else:
                    raise ValueError(
                        f"the system {target_name} does not have an output port named {port_name}."
                    )


            elif isinstance(target[1], list):
                ports_list = []
                for ii, elt_ii in enumerate(target[1]):
                    if isinstance(elt_ii, int):
                        ports_list.append(elt_ii)
                    elif isinstance(elt_ii, str):
                        port_name = elt_ii
                        # Find the system
                        system = None
                        for ii, system_ii in enumerate(system_list):
                            if system_ii.get_name() == target_name:
                                system = system_ii
                                break

                        if system.HasOutputPort(port_name):
                            ports_list += [int(system.GetOutputPort(port_name).get_index())]
                        else:
                            raise ValueError(
                                f"the system {target_name} does not have an output port named {port_name}."
                            )

                    else:
                        raise ValueError(
                            f"the target_list[{ii}] of the tuple is not an integer or a string! " +
                            f"Received {elt_ii} of type {type(elt_ii)}."
                        )
            else:
                raise ValueError(
                    "the second element of the tuple must be either a: \n" +
                    "- an integer\n" +
                    "- a list of integers\n" +
                    "- a string\n" +
                    "- a list of strings\n" +
                    "- None.\n" +
                    f"Received type {type(target[1])}"
                )

        # Create the diagram target object
        dt = DiagramTarget(target_name, ports_list)
        parsed_targets.append(dt)

    return parsed_targets



from brom_drake.control.arms.arm_control_mode import ArmControlMode
from brom_drake.control.arms.end_effector_target import EndEffectorTarget
from brom_drake.control.arms.joint_target import JointTarget
from brom_drake.control.grippers.gripper_target import GripperTarget
from brom_drake.DiagramTarget import DiagramTarget
from brom_drake.DiagramWatcher import DiagramWatcher, DiagramWatcherOptions
from brom_drake.file_manipulation.urdf import (
    drakeify_my_urdf,
    DrakeReadyURDFConverter,
)
from brom_drake.file_manipulation.urdf.drake_ready_urdf_converter.config import (
    DrakeReadyURDFConverterConfig,
    MeshReplacementStrategy,
    MeshReplacementStrategies,
)
from brom_drake.motion_planning.algorithms.rrt import (
    BaseRRTPlannerConfig, BaseRRTPlanner,
    RRTConnectPlannerConfig, RRTConnectPlanner,
)
from brom_drake.motion_planning.systems import (
    PrototypicalPlannerSystem,
    StateOfPlanInMemory,
)
from brom_drake.PortWatcher.port_watcher import PortWatcher
from brom_drake.PortWatcher.port_watcher_options import (
    PortWatcherOptions, FigureNamingConvention,
    PortWatcherPlottingOptions, PortWatcherRawDataOptions,
)
from brom_drake.PortWatcher.port_figure_arrangement import PortFigureArrangement
from brom_drake.productions.ids import ProductionID
from brom_drake.productions.debug.grasping.attempt_grasp.with_puppeteer_wrist import Script as AttemptGraspWithPuppeteerWristScript
from brom_drake.example_helpers import BlockHandlerSystem
from brom_drake.file_manipulation.urdf import DrakeReadyURDFConverter, drakeify_my_urdf
from brom_drake.file_manipulation.urdf.drake_ready_urdf_converter.config import MeshReplacementStrategy
from brom_drake.robots.gripper_type import GripperType
from brom_drake.utils.watcher import add_watcher, add_watcher_and_build, parse_list_of_simplified_targets
from brom_drake.utils.leaf_systems import (
    BoolToVectorSystem,
    EndEffectorWrenchCalculator,
    RigidTransformToVectorSystem,
)
from brom_drake.utils.leaf_systems.rigid_transform_to_vector_system.configuration import Configuration as RigidTransformToVectorSystemConfiguration
from brom_drake.utils.puppetmaker.puppetmaker import Puppetmaker
from brom_drake.utils.puppetmaker.configuration import Configuration as PuppetmakerConfiguration
from brom_drake.utils.puppetmaker.puppet_signature import PuppeteerJointSignature, PuppetSignature
__all__ = [
    'add_watcher',
    'add_watcher_and_build',
    'ArmControlMode',
    'AttemptGraspWithPuppeteerWristScript',
    'BaseRRTPlannerConfig',
    'BaseRRTPlanner',
    'BoolToVectorSystem',
    'DiagramTarget',
    'DiagramWatcher',
    'DiagramWatcherOptions',
    'drakeify_my_urdf', 
    'DrakeReadyURDFConverter',
    'DrakeReadyURDFConverterConfig',
    'EndEffectorTarget',
    'EndEffectorWrenchCalculator',
    'FigureNamingConvention',
    'GripperTarget',
    'GripperType',
    'JointTarget',
    'MeshReplacementStrategy',
    'MeshReplacementStrategies',
    'parse_list_of_simplified_targets',
    'PortWatcher',
    'PortFigureArrangement',
    'PortWatcherOptions',
    'ProductionID',
    'PrototypicalPlannerSystem',
    'PuppeteerJointSignature',
    'Puppetmaker',
    'PuppetmakerConfiguration',
    'PuppetmakerJointSignature',
    'PuppetSignature',
    'RigidTransformToVectorSystem',
    'RigidTransformToVectorSystemConfiguration',
    'RRTConnectPlannerConfig',
    'RRTConnectPlanner',
    'StateOfPlanInMemory',
    'BlockHandlerSystem',
    'DrakeReadyURDFConverter', 'drakeify_my_urdf',
    'PortWatcherPlottingOptions', 'PortWatcherRawDataOptions',
]


from brom_drake.productions.ids import ProductionID
from brom_drake.productions.debug.show_me import ShowMeThisModel
from brom_drake.productions.debug.grasping.attempt_grasp.with_puppeteer_wrist import AttemptGraspWithPuppeteerWrist
from brom_drake.productions.debug.grasping.attempt_grasp.with_puppeteer_wrist.config import Configuration as AttemptGraspWithPuppeteerWristConfiguration
from brom_drake.productions.debug.grasping.attempt_grasp.with_puppeteer_wrist.script import Script as AttemptGraspWithPuppeteerWristScript
from brom_drake.productions.debug.grasping.show_me_this_static_grasp.config import Configuration as ShowMeThisStaticGraspConfiguration
from brom_drake.productions.motion_planning.offline.kinematic import (
    ShelfPlanning1,
    ChemLab1,
)
from brom_drake.productions.motion_planning.offline.dynamic import (
    ChemLab2,
)
from brom_drake.productions.types.base.base import BaseProduction
from brom_drake.productions.types.motion_planning.offline import (
    KinematicMotionPlanningProduction,
    OfflineDynamicMotionPlanningProduction,
)
from brom_drake.productions.types.pick_and_place import (
    MotionPlanningAndGraspingProduction,
)
from brom_drake.productions.pick_and_place import (
    ChemLab3,
)
from brom_drake.productions.debug.grasping.show_me_this_static_grasp.production import (
    ShowMeThisStaticGrasp,
)
from brom_drake.productions.debug.grasping.attempt_grasp.with_static_wrist.production import AttemptGraspWithStaticWrist
from brom_drake.productions.debug.grasping.attempt_grasp.with_static_wrist.config import Configuration as AttemptGraspWithStaticWristConfiguration

__all__ = [
    "AttemptGraspWithStaticWrist",
    "AttemptGraspWithStaticWristConfiguration",
    "AttemptGraspWithPuppeteerWrist",
    "AttemptGraspWithPuppeteerWristConfiguration",
    "AttemptGraspWithPuppeteerWristScript",
    "BaseProduction",
    "ChemLab1",
    "ChemLab2",
    "ChemLab3",
    "KinematicMotionPlanningProduction",
    "MotionPlanningAndGraspingProduction",
    "OfflineDynamicMotionPlanningProduction",
    "ProductionID",
    "ShowMeThisModel",
    "ShowMeThisStaticGrasp",
    "ShowMeThisStaticGraspConfiguration",
    "ShelfPlanning1",
]


from enum import IntEnum
import numpy as np
from pydrake.all import (
    MultibodyPlant,
    RotationMatrix, RigidTransform,
    CoulombFriction,
    HalfSpace, Box,
)

# Class Definition
class GroundShape(IntEnum):
    """
    Description:
        This class defines the different shapes that the ground can have.
    """
    kHalfSpace = 0
    kBox = 1

def AddGround(
    plant: MultibodyPlant,
    shape: GroundShape = GroundShape.kBox,
    z_ground: float = None,
    ground_thickness: float = 0.1,
):
    """
    Description
    -----------
    Add a flat ground with friction

    Arguments
    ---------
    plant: MultibodyPlant
        The plant to add the ground to.
    shape: GroundShape
        The shape of the ground. Defaults to a box.
    z_ground: float
        The height at which the "top" of the ground will be placed.

    """
    # Constants
    transparent_color = np.array([0.5, 0.5, 0.5, 0])
    nontransparent_color = np.array([0.5, 0.5, 0.5, 0.1])

    # Setup

    # Input Processing
    if z_ground is None:
        p_GroundOrigin = [0, 0.0, 0.0]
        R_GroundOrigin = RotationMatrix.MakeXRotation(0.0)
        X_GroundOrigin = RigidTransform(R_GroundOrigin,p_GroundOrigin)
    else:
        p_GroundOrigin = [0, 0.0, z_ground - ground_thickness/2.0]
        R_GroundOrigin = RotationMatrix.MakeXRotation(0.0)
        X_GroundOrigin = RigidTransform(R_GroundOrigin,p_GroundOrigin)

    # Input Processing
    if shape == GroundShape.kHalfSpace:
        shape = HalfSpace()
    elif shape == GroundShape.kBox:
        shape = Box(10, 10, ground_thickness)
    else:
        raise ValueError(f"Invalid shape: {shape}!")  

    # Set Up Ground on Plant
    surface_friction = CoulombFriction(
            static_friction = 0.7,
            dynamic_friction = 0.5)
    plant.RegisterCollisionGeometry(
            plant.world_body(),
            X_GroundOrigin,
            shape,
            "ground_collision",
            surface_friction)
    plant.RegisterVisualGeometry(
            plant.world_body(),
            X_GroundOrigin,
            shape,
            "ground_visual",
            transparent_color)  # transparent


from importlib import resources as impresources
from typing import List, Tuple, Union
from dataclasses import dataclass, field
import networkx as nx
import numpy as np
from pydrake.all import (
    AbstractValue,
    Adder,
    ConstantValueSource,
    GeometryProperties,
    InputPortIndex,
    IllustrationProperties,
    JointActuator,
    ModelInstanceIndex,
    PidController,
    PiecewiseTrajectory,
    PiecewisePolynomial,
    PrismaticJoint,
    RigidBodyFrame,
    RigidTransform,
    SceneGraphInspector,
    TrajectorySource,
)
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import AddMultibodyPlantSceneGraph, MultibodyPlant
from pydrake.systems.framework import DiagramBuilder, Diagram, Context
from pydrake.systems.primitives import ConstantVectorSource, VectorLogSink

# Internal Imports
from brom_drake.PortWatcher.port_watcher_options import FigureNamingConvention
from brom_drake.control.grippers import GripperController, GripperTarget
from brom_drake.file_manipulation.urdf.drakeify import drakeify_my_urdf
from brom_drake.file_manipulation.urdf.shapes.box import BoxDefinition
from brom_drake.file_manipulation.urdf.simple_writer import SimpleShapeURDFDefinition
from brom_drake.robots import find_base_link_name_in, GripperType
from brom_drake import robots
from brom_drake.motion_planning.systems import OpenLoopPlanDispenser
from brom_drake.productions.types.debug import BasicGraspingDebuggingProduction
from brom_drake.productions import ProductionID
from brom_drake.productions.roles.role import Role
from brom_drake.utils import (
    Performer, collision_checking, NetworkXFSM, FlexiblePortSwitch, FSMTransitionCondition,
    FSMOutputDefinition, FSMTransitionConditionType
)
from brom_drake.utils.model_instances import (
    get_name_of_first_body_in_urdf,
    find_number_of_positions_in_welded_model,
)
from brom_drake.productions.debug.show_me.show_me_system import ShowMeSystem
from .config import Configuration
from .phases import AttemptGraspWithStaticWristPhase
from .script import Script as AttemptGraspWithStaticWristScript

class AttemptGraspWithStaticWrist(BasicGraspingDebuggingProduction):
    # Member functions
    def __init__(
        self,
        path_to_object: str,
        gripper_choice: GripperType,
        grasp_joint_positions: np.ndarray,
        X_ObjectTarget: RigidTransform = None,
        meshcat_port_number: int = None,
        script: AttemptGraspWithStaticWristScript = None,
        config: Configuration = Configuration(),
    ):
        # Use the enum to choose the gripper URDF from a set of supported grippers
        if gripper_choice == GripperType.Robotiq_2f_85:
            path_to_gripper = str(
                impresources.files(robots) / "models/robotiq/2f_85_gripper/urdf/robotiq_2f_85.urdf"
            )
        else:
            raise ValueError(f"Gripper type {gripper_choice} not supported for this scene! Create an issue on GitHub if you want it to be!")

        # Handle config
        if meshcat_port_number is not None:
            config.base.meshcat_port_number = meshcat_port_number
        
        if script is not None: # If the user provided a "script" for this production, then use it!
            config.script = script

        # Call the parent constructor
        super().__init__(
            path_to_object=path_to_object,
            path_to_gripper=path_to_gripper,
            X_ObjectGrasp=X_ObjectTarget,
            meshcat_port_number=config.base.meshcat_port_number,
            time_step=config.base.time_step,
            target_body_on_gripper=config.target_body_on_gripper,
            gripper_color=config.gripper_color,
            show_gripper_base_frame=config.show_gripper_base_frame,
            show_collision_geometries=config.show_collision_geometries,
        )

        # This will define the following fields for the production:
        # - path_to_object
        # - path_to_gripper
        # - X_ObjectTarget
        # - meshcat_port_number
        # - time_step
        # - target_body_on_gripper
        # - gripper_color
        # - show_gripper_base_frame
        # - show_collision_geometries
        # - plant

        # TODO(kwesi): Figure out a better way to handle the initial joint positions
        # Create INITIAL joint position array
        n_gripper_positions = find_number_of_positions_in_welded_model(self.path_to_gripper)
        # if initial_gripper_joint_positions is None:
        #     initial_gripper_joint_positions = [0.0] * n_gripper_positions
        self.initial_gripper_joint_positions = np.zeros(grasp_joint_positions.shape)# initial_gripper_joint_positions

        # Save target joint position array
        # assert len(grasp_joint_positions) != n_gripper_positions, \
        #     f"Expected for the \"grasp_joint_positions\" array to contain {n_gripper_positions} values (the number of positions)" + \
        #     f"in the model, but received length {len(grasp_joint_positions)} array."
        
        self.grasp_joint_positions = grasp_joint_positions
        self.config = config
        
        # Add Name to plantPlant and Scene Graph for easy simulation
        self.plant.set_name("DemonstrateStaticGrasp_plant")

        # Create show me system for holding object in place
        self.show_me_system = None

        self.floor_model_index = None
        self.floor_actuator = None
        self.floor_shape = None

        # Create an executive/brain
        self.executive = self.create_executive_system() # The executive system that coordinates ALL the systems

    def add_cast_and_build(
        self,
        cast: Tuple[Role, Performer] = [],
        figure_naming_convention: FigureNamingConvention = FigureNamingConvention.kHierarchical,
    ) -> Tuple[Diagram, Context]:
        super().add_cast_and_build(cast, with_watcher=True, figure_naming_convention=figure_naming_convention)

        return self.diagram, self.diagram_context

    def add_floor_to_plant(
        self,
        floor_mass: float = 100.0,
        plant: MultibodyPlant = None,
    ) -> Tuple[
        float, List[float], ModelInstanceIndex, JointActuator
    ]:
        """
        Description
        -----------
        This method will add a floor to the plant.
        """
        # Setup
        if plant is None:
            plant: MultibodyPlant = self.plant

        # Create a box urdf for the floor
        floor_shape = [10.0, 10.0, 0.1] # Length, Width, Height
        floor_geometry_defn = BoxDefinition(size=floor_shape)

        # Create a urdf for the floor
        floor_urdf_defn = SimpleShapeURDFDefinition(
            name="floor",
            shape=floor_geometry_defn,
            mass=floor_mass,
        ) 
        floor_urdf = floor_urdf_defn.write_to_file()

        # Add the floor to the plant
        floor_model_idcs = Parser(plant=plant).AddModels(floor_urdf)
        floor_model_index = floor_model_idcs[0]

        # Create joint + actuator for the floor
        floor_joint = plant.AddJoint(
            PrismaticJoint(
                name="floor_joint",
                frame_on_parent=plant.world_frame(),
                frame_on_child=plant.GetFrameByName(get_name_of_first_body_in_urdf(floor_urdf)),
                axis=[0, 0, 1],
            )
        )

        floor_actuator = plant.AddJointActuator(
            "floor_elevation_actuator",
            floor_joint,
        )

        return floor_mass, floor_shape, floor_model_index, floor_actuator

    def add_supporting_cast(self):
        """
        Description
        -----------
        This method will add:
        - The user's object model to the builder.
        - The user's gripper model to the builder.
        - The gripper triad to the builder.
        """
        # Setup
        plant: MultibodyPlant = self.plant

        # Add the object to the builder
        self.add_manipuland_to_plant()

        # Add the gripper to the builder
        X_WorldGripper = self.find_X_WorldGripper(
            X_ObjectGripper=self.X_ObjectGripper,
            target_frame_name=self.target_body_name_on_gripper,
            desired_joint_positions=self.initial_gripper_joint_positions,
        )
        self.add_gripper_to_plant(
            and_weld_to=plant.world_frame(),
            with_X_WorldGripper=X_WorldGripper,
        )

        # Add the floor
        floor_mass, self.floor_shape, self.floor_model_index, self.floor_actuator = self.add_floor_to_plant()

        # Connect the plant to the meshcat, if requested
        if self.meshcat_port_number is not None:
            self.connect_to_meshcat()

        # Finalize the plant
        self.plant.Finalize()

        # Add controllers for gripper AND floor
        self.add_gripper_controller_and_connect()
        self.add_floor_controller_and_connect(floor_mass)

        # Create defaults for plant
        self.set_plant_defaults()

    def add_gripper_trajectory_system(self) -> TrajectorySource:
        # Setup
        builder: DiagramBuilder = self.builder
        script: AttemptGraspWithStaticWristScript = self.config.script

        # Collect gripper initial and final positions
        p_gripper0 = self.initial_gripper_joint_positions
        p_gripper_final = self.grasp_joint_positions

        # Find time to start closing the gripper from the script
        t_gripper_close_start = script.start_time_of_phase(
            phase=AttemptGraspWithStaticWristPhase.kGripperClosing
        )
        t_object_settling_in_grasp_start = script.start_time_of_phase(
            phase=AttemptGraspWithStaticWristPhase.kObjectSettlingInGrasp
        )

        # Create a piecewise trajectory for the gripper to follow'
        joint_position_samples = []
        joint_position_samples.append(p_gripper0.reshape(-1, 1).tolist()) # This wild compression is needed because of bugs in Drake. :'(
        joint_position_samples.append(p_gripper0.reshape(-1, 1).tolist())
        joint_position_samples.append(p_gripper_final.reshape(-1, 1).tolist())
        gripper_trajectory = PiecewisePolynomial.FirstOrderHold(
            samples=joint_position_samples,
            breaks=[0.0, t_gripper_close_start, t_object_settling_in_grasp_start],
        )

        # Add a trajectory source object to the builder
        gripper_trajectory_source = builder.AddSystem(
            TrajectorySource(gripper_trajectory)
        )
        gripper_trajectory_source.set_name("[Gripper] Trajectory Source")
        return gripper_trajectory_source


    def add_gripper_controller_and_connect(self):
        # Setup
        builder: DiagramBuilder = self.builder
        plant: MultibodyPlant = self.plant

        # Define System that dispatches the gripper trajectory
        gripper_trajectory_source = self.add_gripper_trajectory_system()

        # Create a simple piece of logic to turn on/off the gripper (i.e., to close it)
        gripper_controller = builder.AddSystem(
            GripperController(
                GripperType.Robotiq_2f_85,
                Kp=1_000.0 * np.eye(2),
            ),
        )
        gripper_controller.set_name("[Gripper] Gripper Controller")

        # Connect the gripper controller to the gripper
        builder.Connect(
            gripper_controller.GetOutputPort("applied_gripper_torque"),
            plant.get_actuation_input_port(self.gripper_model_index),
        )

        # Connect plan to the controller
        builder.Connect(
            gripper_trajectory_source.get_output_port(),
            gripper_controller.GetInputPort("gripper_target"),
        )

        # Create a signal that tells the controller to use position control
        gripper_position_control_source = builder.AddSystem(
            ConstantValueSource(
                AbstractValue.Make(GripperTarget.kPosition),
            )
        )
        gripper_position_control_source.set_name("[Gripper] Target Type Signal")
        builder.Connect(
            gripper_position_control_source.get_output_port(),
            gripper_controller.GetInputPort("gripper_target_type"),
        )

        # Connect the plant to the gripper controller
        builder.Connect(
            plant.get_state_output_port(self.gripper_model_index),
            gripper_controller.GetInputPort("gripper_state"),
        )


    def add_floor_controller_and_connect(self, floor_mass: float):
        # Setup
        plant: MultibodyPlant = self.plant

        # Find the z position of the floor
        z_floor = self.find_floor_z_via_line_search()

        # Connect a PID Controller to the floor actuator
        kp = np.array([[floor_mass * 9.81 * 1.0e0]]) #Idk how I'm picking this number.
        ki = np.zeros(kp.shape) # 0.1 * np.sqrt(kp)
        kd = 4.0 * np.sqrt(kp)
        floor_controller = self.builder.AddSystem(
            PidController(kp=kp, ki=ki, kd=kd),
        )
        floor_controller.set_name("[Floor] PID Controller")

        # Create a feedforward term that we will add to the PID controller's
        # output to keep the floor at a constant height
        # (i.e., the weight of the object)
        feedforward_term = self.builder.AddSystem(
            ConstantVectorSource(np.array([[floor_mass * 9.81]])),  # Weight of the object
        )
        feedforward_term.set_name("[Floor] Feedforward")

        # Summer to combine the PID output and the feedforward term
        summer = self.builder.AddSystem(
            Adder(2, 1),
        )

        self.builder.Connect(
            floor_controller.get_output_port(0),  # PID output
            summer.get_input_port(0),  # summer input
        )

        self.builder.Connect(
            feedforward_term.get_output_port(0),  # Weight of the object
            summer.get_input_port(1),  # Feedforward term
        )

        # Connect the target height to the PID controller
        target_height_source = self.create_floor_trajectory_source(z_floor=z_floor)

        # Connect the source to the floor actuator
        self.builder.Connect(
            summer.get_output_port(0),
            plant.get_actuation_input_port(self.floor_model_index),
        )

        # Connect the target height to the PID controller
        self.builder.Connect(
            target_height_source.get_output_port(0),
            floor_controller.get_input_port_desired_state(),
        )

        # Connect the current height of the floor to the PID controller
        self.builder.Connect(
            plant.get_state_output_port(self.floor_model_index),
            floor_controller.get_input_port_estimated_state(),
        )

    def create_executive_system(self) -> NetworkXFSM:
        # Setup
        builder: DiagramBuilder = self.builder
        script: AttemptGraspWithStaticWristScript = self.config.script

        # Create the NetworkXFSM from the graph
        executive = builder.AddSystem(script.to_fsm())

        return executive

    def create_floor_trajectory_source(
        self,
        z_floor: float
    ) -> Tuple[FlexiblePortSwitch]:
        # Setup
        builder: DiagramBuilder = self.builder

        #Create a proper trajectory source for the floor so that it slowly falls away
        floor_target_height_source = builder.AddSystem(
            OpenLoopPlanDispenser(2, speed=0.25),
        )

        # Create the trajectory that will be run
        simple_plan = np.array([
            [z_floor, 0.0],
            [z_floor - 10.0, 0.0]
        ])
        floor_trajectory_source = builder.AddSystem(
            ConstantValueSource(AbstractValue.Make(simple_plan))
        )
        
        builder.Connect(
            floor_trajectory_source.get_output_port(),
            floor_target_height_source.GetInputPort("plan"),
        )

        # Connect timer to the plan dispenser
        builder.Connect(
            self.executive.GetOutputPort("start_floor"),
            floor_target_height_source.GetInputPort("plan_ready"),  # Trigger input
        )

        return floor_target_height_source

    def find_floor_z_via_bounding_box(
        self,
        debug_flag: bool = True,
    ) -> float:
        """
        Description
        -----------
        This method will find the transform needed to put the floor at so that
        the object sits "just" on top of it.

        WARNING: This method does not work and I don't know why. (i.e., sometimes
        the floor is in collision with the object)
        """
        # Setup
        shadow_builder = DiagramBuilder()

        # Create shadow plant + scene graph
        shadow_plant, shadow_scene_graph = AddMultibodyPlantSceneGraph(
            shadow_builder,
            time_step=self.time_step,
        )
        shadow_plant.set_name("DemonstrateStaticGrasp_shadow_plant")
        shadow_scene_graph.set_name("DemonstrateStaticGrasp_shadow_scene_graph")

        # Add the object to the shadow plant
        model_idcs = Parser(plant=shadow_plant).AddModels(self.path_to_object)
        assert len(model_idcs) == 1, f"Only one model should be added; received {len(model_idcs)}"
        model_idx = model_idcs[0]
        model_name = shadow_plant.GetModelInstanceName(model_idx)

        # Weld the base link to the world frame
        # manipuland_body_idcs = shadow_plant.GetBodyIndices(model_idx)
        # assert len(manipuland_body_idcs) > 0, \
        #     f"Expected at least one body in the manipuland; received {len(manipuland_body_idcs)}"
        # manipuland_body = shadow_plant.get_body(manipuland_body_idcs[0])
        # frame0 = manipuland_body.body_frame()
        # shadow_plant.WeldFrames(
        #     shadow_plant.world_frame(),
        #     frame0,
        # )

        # Finalize the shadow plant
        shadow_plant.Finalize()

        # Search through the geometries in the scene and find the manipuland's
        # geometry. Attempt to compute the bounding box around it.
        shadow_sg_inspector: SceneGraphInspector = shadow_scene_graph.model_inspector()
        all_geometries = shadow_scene_graph.model_inspector().GetAllGeometryIds()
        objects_minimum_z = 1_000.0
        for ii, geometry_ii in enumerate(all_geometries):
            # Announce geometry value:
            if debug_flag:
                print(f"Geometry ID {ii}:")
                print(f"- Raw value: {geometry_ii}")

            # Get shape of the geometry
            shape_ii = shadow_sg_inspector.GetShape(geometry_ii)
            if debug_flag:
                print(f"- Shape {ii}:")
                print(f"  + Raw value: {shape_ii}")
                print(f"  + Type: {type(shape_ii)}")

            # Get the convex hull
            ch_ii = shape_ii.GetConvexHull()
            if debug_flag:
                print(f"- Convex Hull {ii}:")
                print(f"  + Raw value: {ch_ii}")

            bb_ii = ch_ii.CalcBoundingBox()
            if debug_flag:
                print(f"- Bounding Box {ii}:")
                print(f"  + Raw value: {bb_ii}")
            lower_left, upper_right = bb_ii

            if lower_left[2] < objects_minimum_z:
                objects_minimum_z = lower_left[2]
                print(f"  + New z_out: {objects_minimum_z}")
        
        # If we didn't find any geometry, raise an error
        if objects_minimum_z == 1_000.0:
            raise ValueError("Unable to find geometry for the manipuland.")
        
        # If we found a geometry, then we need to set the floor height
        # using the known shape of the table

        return objects_minimum_z - self.floor_shape[2]*0.5  # Subtract the height of the floor

    def find_floor_z_via_line_search(
        self,
        debug_flag: bool = False,
    ) -> float:
        """
        Description
        -----------
        Use bounding box method to generate an initial guess for the floor height
        and then perform a line search to find the height that actually
        is collision free.
        """
        # Setup
        z_floor = self.find_floor_z_via_bounding_box(debug_flag=debug_flag)
        floor_shape = self.floor_shape

        # Check for collisions at the initial guess
        collision_detected = self.floor_z_creates_collisions(
            floor_z_translation=z_floor,
            debug_flag=debug_flag,
        )
        if not collision_detected:
            # If no collision detected, return the initial guess
            return z_floor
        
        # If we have a collision, then we need to perform a line search
        # to find the height that is collision free
        
        # Start with a small step size
        step_size = floor_shape[2] * 0.25  # 25% of the floor height

        # Perform a line search to find the height that is collision free
        candidate_z_floor = z_floor - step_size
        while True:
            # Check for collisions at the current height
            collision_detected = self.floor_z_creates_collisions(
                floor_z_translation=candidate_z_floor,
                debug_flag=debug_flag,
            )
            if not collision_detected:
                # If no collision detected, return the current height
                print(f"Found collision-free height: {candidate_z_floor}")
                return candidate_z_floor
            

            print(f"Collision detected at height: {candidate_z_floor}")
            # If we have a collision, then increase the height
            candidate_z_floor -= step_size

            # If we exceed a certain height, break out of the loop
            if candidate_z_floor < -10.0:
                break        
                
        raise ValueError("Unable to find a collision-free height for the floor.")

    def find_X_WorldGripper(
        self,
        X_ObjectGripper: RigidTransform,
        target_frame_name: str,
        desired_joint_positions: List[float],
    ) -> RigidTransform:
        """
        Description
        -----------
        This method will return the transform from the world frame to the gripper frame.
        """
        # Setup
        desired_joint_positions = np.array(desired_joint_positions)

        # Create shadow plant containing just the gripper
        shadow_plant = MultibodyPlant(self.time_step)
        model_idcs = Parser(plant=shadow_plant).AddModels(self.path_to_gripper)
        shadow_model_idx = model_idcs[0]
        
        # Get the first body on the gripper
        name_of_first_body_in_gripper = get_name_of_first_body_in_urdf(self.path_to_gripper)

        # Weld the base link to the world frame
        shadow_plant.WeldFrames(
            shadow_plant.world_frame(),
            shadow_plant.GetFrameByName(name_of_first_body_in_gripper),
        )

        # Finalize Plant
        shadow_plant.Finalize()

        shadow_builder = DiagramBuilder()
        shadow_builder.AddSystem(shadow_plant)
        diagram = shadow_builder.Build()
        shadow_diagram_context = diagram.CreateDefaultContext()

        # Set the joint positions
        # shadow_plant.SetPositions(
        #     shadow_plant.GetMyContextFromRoot(shadow_diagram_context),
        #     shadow_model_idx,
        #     desired_joint_positions,
        # )

        # Get the transform of the gripper's base in the static world frame
        X_GripperBase_GripperTarget = shadow_plant.EvalBodyPoseInWorld(
            shadow_plant.GetMyContextFromRoot(shadow_diagram_context),
            shadow_plant.GetBodyByName(target_frame_name),
        )

        X_Object_GripperTarget = X_ObjectGripper
        X_WorldObject = RigidTransform.Identity()
        X_Object_GripperBase = X_Object_GripperTarget.multiply(
            X_GripperBase_GripperTarget.inverse(),
        )

        return X_WorldObject.multiply(X_Object_GripperBase)

    def floor_z_creates_collisions(
        self,
        floor_z_translation: float = 0.0,
        debug_flag: bool = False,
    ) -> bool:
        # Setup
        shadow_builder = DiagramBuilder()

        # Create shadow plant + scene graph
        shadow_plant, shadow_scene_graph = AddMultibodyPlantSceneGraph(
            shadow_builder,
            time_step=self.time_step,
        )
        shadow_plant.set_name("DemonstrateStaticGrasp_shadow_plant")
        shadow_scene_graph.set_name("DemonstrateStaticGrasp_shadow_scene_graph")

        # Add the object to the shadow plant
        model_idcs = Parser(plant=shadow_plant).AddModels(self.path_to_object)
        assert len(model_idcs) == 1, f"Only one model should be added; received {len(model_idcs)}"
        model_idx = model_idcs[0]
        model_name = shadow_plant.GetModelInstanceName(model_idx)

        # # Add the gripper to the shadow plant
        # temp_idcs = Parser(plant=shadow_plant).AddModels(self.path_to_gripper)
        # assert len(temp_idcs) == 1, f"Only one model should be added; received {len(temp_idcs)}"
        # shadow_gripper_model_index = temp_idcs[0]
        # shadow_gripper_model_name = shadow_plant.GetModelInstanceName(shadow_gripper_model_index)

        # Add the floor to the shadow plant
        _, _, shadow_floor_model_index, _ = self.add_floor_to_plant(plant=shadow_plant)

        # Finalize the shadow plant
        shadow_plant.Finalize()

        # Create diagram and set position of the floor to be floor_z_translation
        shadow_diagram = shadow_builder.Build()
        shadow_diagram_context = shadow_diagram.CreateDefaultContext()

        shadow_plant.SetPositions(
            shadow_plant.GetMyContextFromRoot(shadow_diagram_context),
            shadow_floor_model_index,
            np.array([floor_z_translation]),
        )

        return collision_checking.using_point_pair_penetration(
            shadow_plant, shadow_scene_graph, shadow_diagram_context,
        )

    @property
    def id(self):
        return ProductionID.kAttemptGraspWithStaticWrist

    def set_plant_defaults(self):
        """
        Description
        -----------
        This method will set the default state of the plant to the desired joint positions.
        """
        # Setup
        plant: MultibodyPlant = self.plant

        # Find the z position of the floor
        z_floor = self.find_floor_z_via_line_search()

        # Set the default state of the floor to be at the desired height
        plant.SetDefaultPositions(
            self.floor_model_index,
            np.array([z_floor]),
        )

        n_gripper_positions = find_number_of_positions_in_welded_model(self.path_to_gripper)
        plant.SetDefaultPositions(
            self.gripper_model_index,
            np.zeros((n_gripper_positions,)),
        )

    @property
    def suggested_roles(self) -> List[Role]:
        return []
    
    


"""
Description
-----------
    This file contains the DiagramWatcherOptions class.
"""

from dataclasses import dataclass

# Internal Imports
from brom_drake.directories import DEFAULT_WATCHER_DIR
from brom_drake.PortWatcher.port_watcher_options import PortWatcherOptions, PortWatcherPlottingOptions, PortWatcherRawDataOptions

@dataclass(frozen=True)
class SuppressDiagramWatcherRules:
    """
    *Description*
    
    A signal that defines which types of messages to suppress when
    using the DiagramWatcher.

    *Attributes*

    during_port_watcher_connection: bool
        If True, suppress log messages that occur during the connection
        of PortWatchers to the DiagramWatcher.
    """
    # The list of rules to suppress the DiagramWatcher
    during_port_watcher_connection: bool = False

@dataclass
class DiagramWatcherOptions:
    """
    *Description*
    
    This class contains the options that configure the following properties of the DiagramWatcher:
    - Where it saves its data (plots, raw data, etc.)
    - the format that it uses to save its data/plots
    - what types of messages to suppress during its operation

    .. tip::

        The DiagramWatcherOptions class has useful defaults defined for ALL
        of its attributes, so in most cases, the user will not need to modify any of them.
        You can simply create a DiagramWatcherOptions object with no arguments
        and it will work out of the box. (e.g., `options = DiagramWatcherOptions()`)

    *Attributes*

    base_directory: str, optional
        The base directory where the watcher will save any data (e.g., plots, or dataframes)
        during its operation as well as logs.

    plotting_options: PortWatcherPlottingOptions, optional
        The options that configure how the PortWatcher will save its plots.

    raw_data_options: PortWatcherRawDataOptions, optional
        The options that configure how the PortWatcher will save its raw data.

    hide_messages: SuppressDiagramWatcherRules, optional
        The rules that define which types of messages to suppress during the operation of the DiagramWatcher.
    """
    # The base directory where the watcher will save the data
    base_directory: str = DEFAULT_WATCHER_DIR

    # Options Used for Each PortWatcher
    plotting_options: PortWatcherPlottingOptions = PortWatcherPlottingOptions()
    raw_data_options: PortWatcherRawDataOptions = PortWatcherRawDataOptions()

    # Flags for hiding the outputs of the DiagramWatcher
    hide_messages: SuppressDiagramWatcherRules = SuppressDiagramWatcherRules()

    def to_port_watcher_options(self) -> PortWatcherOptions:
        """
        *Description*
        
        This function converts the DiagramWatcherOptions to a PortWatcherOptions object.

        *Returns*
        
        PortWatcherOptions
            The PortWatcherOptions object.
        """
        return PortWatcherOptions(
            plotting=self.plotting_options,
            raw_data=self.raw_data_options,
        )
    


from dataclasses import dataclass
import xml.etree.ElementTree as ET

# Internal Import
from .shape_definition import ShapeDefinition, ShapeEnum

@dataclass
class CylinderDefinition(ShapeDefinition):
    """
    *Description*
    
    Defines a cylinder shape as per URDF specifications.

    *Attributes*

    radius: float
        The radius of the cylinder.

    length: float
        The length of the cylinder along its central axis.

    """
    radius: float
    length: float

    def add_geometry_to_element(self, target_element: ET.Element) -> ET.Element:
        """
        *Description*
        
        Adds the cylinder geometry as a sub-element to the given target element.
        
        This method is an implementation of the abstract method defined in the
        ShapeDefinition base class.

        *Parameters*

        target_element: xml.etree.ElementTree.Element
            The target XML element to which the cylinder geometry will be added.

        *Returns*

        xml.etree.ElementTree.Element
            The element `target_element` with the cylinder geometry added as a sub-element.
        """
        return ET.SubElement(
            target_element,
            "cylinder",
            {"radius": f"{self.radius}", "length": f"{self.length}"},
        )

    @property
    def type(self) -> ShapeEnum:
        """
        *Description*

        Returns the ShapeEnum type corresponding to a cylinder.

        This method is an implementation of the abstract method defined in the
        ShapeDefinition base class.

        *Returns*

        ShapeEnum
            The ShapeEnum type for a cylinder.
        """

        return ShapeEnum.kCylinder


from pydrake.geometry import SceneGraph
from pydrake.systems.primitives import (
    VectorLogSink, ConstantVectorSource,
)

# List of system types that should be ineligible for diagram watching
INELIGIBLE_SYSTEM_TYPES = [SceneGraph, VectorLogSink, ConstantVectorSource]


from enum import IntEnum

class DispenserInternalState(IntEnum):
    kReady = 0
    kPlanSet = 1

class DispenserTransitionRequest(IntEnum):
    kNone = 0               # Do not request to a new internal state
    kRequestSavePlan = 1    # Request to save the current plan into memory; Requires that the planner be in the kReady state.
    kRequestReset = 2       # Request to reset the dispenser; Requires that the planner be in the kPlanSet state.


from brom_drake.DiagramWatcher.DiagramWatcher import DiagramWatcher
from brom_drake.DiagramWatcher.errors import UnrecognizedTargetError
from brom_drake.DiagramWatcher.diagram_watcher_options import DiagramWatcherOptions

__all__ = [
    'DiagramWatcher',
    'DiagramWatcherOptions',
    'UnrecognizedTargetError',
]



from brom_drake.productions.types.base.base import BaseProduction
from brom_drake.productions.types.motion_planning.offline import (
    KinematicMotionPlanningProduction,
    OfflineDynamicMotionPlanningProduction,
)
from brom_drake.productions.types.pick_and_place import (
    MotionPlanningAndGraspingProduction,
)

__all__ = [
    "BaseProduction",
    "KinematicMotionPlanningProduction",
    "MotionPlanningAndGraspingProduction",
    "OfflineDynamicMotionPlanningProduction",
]


"""
PortWatcher.py
Description:

    This file defines the PortWatcher class. This class is used to watch the ports of a
    diagram.
"""
import logging
from pathlib import Path
from typing import Dict
import numpy as np
import os

from pydrake.multibody.plant import MultibodyPlant
from pydrake.systems.framework import OutputPort, PortDataType, DiagramBuilder, LeafSystem
from pydrake.systems.primitives import LogVectorOutput, VectorLogSink
from pydrake.systems.framework import Context

# Internal Imports
from brom_drake.directories import DEFAULT_BROM_DIR
from .file_manager import PortWatcherFileManager
from brom_drake.PortWatcher.support_types import assert_port_is_supported
from .port_watcher_options import (
    PortWatcherOptions, FigureNamingConvention,
    PortWatcherPlottingOptions, PortWatcherRawDataOptions,
)
from .plotter import PortWatcherPlotter
from brom_drake.systems.abstract_list_selection_system import AbstractListSelectionSystem
from brom_drake.utils import (
    BoolToVectorSystem,
    RigidTransformToVectorSystem,
)
from brom_drake.utils.plant import get_all_associated_body_indices_in_plant
from brom_drake.utils.type_checking import is_rigid_transform

OutputPortNameLike = str

class PortWatcher:
    """
    *Description*

    The real workhorse of the :py:class:`DiagramWatcher<brom_drake.DiagramWatcher.DiagramWatcher.DiagramWatcher>` class.
    This class adds the elements to the drake diagram that will monitor a given
    system's output port (**output_port**), if possible.
    """
    def __init__(
        self,
        output_port: OutputPort,
        builder: DiagramBuilder,
        python_logger: logging.Logger,
        logger_name: str = None,
        options: PortWatcherOptions = PortWatcherOptions(),
        base_watcher_dir: str = DEFAULT_BROM_DIR
    ):
        """
        *Description*
        
        This class is used to watch a single port of a system.

        *Parameters*
        
        output_port: OutputPort
            The output port that will be watched.

        builder: DiagramBuilder
            The diagram builder that is used to create the logger.

        logger_name: str
            The name of the logger.

        options: PortWatcherOptions
            The options that are used to configure the watcher.

        plot_dir: str
            The directory where the plots will be saved.

        raw_data_dir: str
            The directory where the raw data will be saved.
        """
        # Setup
        self.options = options
        self.port = output_port
        self.data = {}
        self.plot_handles = {}
        self.plot_handles = None
        self.file_manager = PortWatcherFileManager(
            base_directory=Path(base_watcher_dir),
            plotting_options=self.options.plotting,
            raw_data_file_format=self.options.raw_data.file_format,
        )
        self.python_logger = python_logger

        # Set up directories
        os.makedirs(self.file_manager.plot_dir, exist_ok=True)
        os.makedirs(self.file_manager.raw_data_dir, exist_ok=True)

        # Input Processing
        assert_port_is_supported(self.port)

        # Preparing LogVectorSink
        self._drake_vector_logs: Dict[OutputPortNameLike, VectorLogSink] = {}
        self.prepare_vector_logs(builder)

        # Prepare optional members
        self.plotter = None
        if self.options.plotting.save_to_file:
            self.plotter = PortWatcherPlotter(
                port=self.port,
                python_logger=self.python_logger,
                plotting_options=self.options.plotting,
                file_manager=self.file_manager,
            )

    def get_vector_log_sink(self, with_index: int = None, with_output_port_name: str = None) -> VectorLogSink:
        """
        *Description*

        Returns the VectorLogSink corresponding to the given index or output port name.
        By default, this returns the first VectorLogSink in the internal dictionary.

        *Parameters*
        
        with_index: int, optional
            The index of the VectorLogSink to return.
            By default, None.

        with_output_port_name: str, optional
            The name of the output port of the VectorLogSink to return.
            By default, None.
            
        *Returns*
        
        vector_log_sink: VectorLogSink
            The VectorLogSink corresponding to the given index or output port name.
        """
        # Default case
        if with_index is None and with_output_port_name is None:
            all_vector_logs = list(self._drake_vector_logs.values())
            return all_vector_logs[0]

        # If either parameter is provided, use it to select the log
        if with_index is not None:
            all_vector_logs = list(self._drake_vector_logs.values())
            return all_vector_logs[with_index]
        elif with_output_port_name is not None:
            return self._drake_vector_logs[with_output_port_name]
        else:
            raise ValueError(
                "Either with_index or with_output_port_name must be provided."
            )

    def name_vector_log_sink(
        self,
        current_output_port: OutputPort,
    ):
        """
        *Description*

        Provides a name to the VectorLogSink stored in `self.drake_vector_logs`
        that corresponds to the given system and output_port.
        """
        # Setup
        original_output_port = self.port
        system = original_output_port.get_system()

        # Name the VectorLogSink        
        name = f"PortWatcher_{system.get_name()}_{original_output_port.get_name()}"
        if current_output_port.get_name() != original_output_port.get_name():
            name += f"_{current_output_port.get_name()}"

        self._drake_vector_logs[current_output_port.get_name()].set_name(name)

    def prepare_vector_log_for_rigid_transform_port(
        self,
        current_output_port: OutputPort,
        builder: DiagramBuilder,
    ):
        """
        *Description*

        Adds to the current Drake Diagram a VectorLogSink (and helper system) to measure the value of the RigidTransform
        value coming from the **PortWatcher**'s target output port.

        *Parameters*
        
        builder: DiagramBuilder
            The diagram builder that is used to create the VectorLogSink.
        """
        # Determine the name of the system that converts
        # RigidTransform objects to vectors
        converter_name = f"RigidTransformToVectorSystem_{self.safe_system_name(current_output_port.get_system())}"

        # First, let's create an intermediate system
        # that will convert the RigidTransform to a vector that is easily logged
        converter_system = builder.AddNamedSystem(
            system=RigidTransformToVectorSystem(),
            name=converter_name,
        )

        # Connect the system to the port
        builder.Connect(
            current_output_port,
            converter_system.get_input_port(),
        )

        # Then connect the output of the converter to a VectorLogSink
        self._drake_vector_logs[current_output_port.get_name()] = \
            LogVectorOutput(
                converter_system.get_output_port(),
                builder,
            )
        
        # And finally, name the vector log sink (must have unique names to compile the diagram)
        self.name_vector_log_sink(current_output_port=current_output_port)

    def prepare_vector_log_for_abstract_valued_port(
        self,
        current_output_port: OutputPort,
        builder: DiagramBuilder,
    ):
        """
        *Description*

        .. note::

            While this is usually called with the port ``self.port``,
            it can be called with other ports (as is done with list[T] inputs).
        """
        # Collect the system and output port (their names will be used in the default name)
        system: LeafSystem = current_output_port.get_system()

        # Check to see if the port contains a value of type
        # - RigidTransform, or
        # - bool
        example_allocation = current_output_port.Allocate()
        example_value = example_allocation.get_value()

        # Address the case of list type inputs first
        if type(example_value) is list:
            # A list input should NEVER have zero length
            assert len(example_value) > 0, \
                f"All list output ports should contain at least 1 example allocation value; received 0 from port {current_output_port.get_name()}"

            # If the value is a list, then we will use
            # a special system to select each element of the list
            # and then recursively call this function.

            for idx in range(len(example_value)):
                output_port_name = f"element_{idx}_out"

                # If this is a very specific port (body_poses),
                # then propose a very specific name for the output port
                if current_output_port.get_name() == "body_poses":
                    # TODO(Kwesi): Include test to see if
                    # this system is a multibody plant
                    system_is_multibody_plant = type(current_output_port.get_system()) is MultibodyPlant
                    if system_is_multibody_plant:
                        # Extract all body names
                        system_as_multibody_plant: MultibodyPlant = current_output_port.get_system()
                        body_indices = get_all_associated_body_indices_in_plant(system_as_multibody_plant)
                        current_body_name = system_as_multibody_plant.get_body(body_index=body_indices[idx]).name()

                        output_port_name = f"{current_body_name}"

                # Create selection system for this iteration
                selection_i = builder.AddNamedSystem(
                    system = AbstractListSelectionSystem(
                        idx,
                        output_type=type(example_value[0]),
                        output_port_name=output_port_name
                    ),
                    name = f"System {system.get_name()}'s Port \"{current_output_port.get_name()}\" Element #{idx}"
                )

                # Connect selection system to target port
                builder.Connect(
                    current_output_port,
                    selection_i.get_input_port()
                )

                # Use recursion to assign the proper logger to the output of selection_i
                self.prepare_vector_log_for_abstract_valued_port(
                    current_output_port=selection_i.get_output_port(),
                    builder=builder,
                )

            # After the recursion, we should be able to return
            return

        # Non-list types
        if is_rigid_transform(example_value):
            self.prepare_vector_log_for_rigid_transform_port(
                current_output_port, builder,
            )
            self.name_vector_log_sink(current_output_port=current_output_port)

        elif type(example_value) == bool:
            # If the value is a boolean,
            # then we must create an intermediate BoolToVectorSystem
            # that will convert the boolean to a vector.
            converter_system = builder.AddSystem(
                BoolToVectorSystem()
            )

            # Connect the system to the port
            builder.Connect(
                current_output_port,
                converter_system.get_input_port(),
            )

            # Then connect the output of the converter to a logger
            self._drake_vector_logs[current_output_port.get_name()] = LogVectorOutput(
                converter_system.get_output_port(),
                builder,
            )
            self.name_vector_log_sink(current_output_port=current_output_port)
        
        else:
            raise NotImplementedError(
                f"PortWatcher does not support the type of data ({type(example_value)}) contained in the port."
            )

    def prepare_vector_logs(self, builder: DiagramBuilder):
        """
        *Description*
        
        Prepares any VectorLogSink's needed to measure the current port.
        In most cases, only one VectorLogSink is needed.
        
        *Parameters*
        
        builder: DiagramBuilder
            The diagram builder that is used to create the logger.
        
        """
        # Create ALL of the VectorLogSink's needed
        # for each type of data in the port
        if self.port.get_data_type() == PortDataType.kVectorValued:
            self._drake_vector_logs[self.port.get_name()] = LogVectorOutput(self.port, builder)
            self.name_vector_log_sink(current_output_port=self.port)
        else:
            # Port must be abstract valued
            self.prepare_vector_log_for_abstract_valued_port(self.port, builder)

        # Announce the preparation of the VectorLogSink's
        # with the logger
        for target_port_name, log_sink in self._drake_vector_logs.items():
            self.python_logger.info(
                f"PortWatcher prepared VectorLogSink for system \"{self.port.get_system().get_name()}\"'s "
                f"port \"{target_port_name}\".\n" + \
                f" - It's name is \"{log_sink.get_name()}\"."
            )
    
    def safe_system_name(self, system: LeafSystem = None) -> str:
        """
        *Description*
        
        Returns a safe name for the system.

        *Returns*
        
        name: str
            System's name
        """
        # Setup
        if system is None:
            system = self.port.get_system()

        out = system.get_name()

        # First, let's check to see how many "/" exist in the name
        slash_occurences = [i for i, letter in enumerate(out) if letter == "/"]
        if len(slash_occurences) > 0:
            out = out[slash_occurences[-1] + 1:]  # truncrate string based on the last slash

        # Second, replace all spaces with underscores
        out = out.replace(" ", "_")

        return out

    def save_all_figures(self, diagram_context: Context):
        """
        *Description*

        Saves all of the figures that the PortWatcher should according to its options.
        This may generate one figure for each dimension of the output port being watched,
        or just one figure for the entire port.

        *Parameters*
        
        diagram_context: Context
            The context of the diagram.
        """
        # Test to see if we have a plotter
        assert self.plotter is not None, \
            "Cannot save figures because no plotter was created for this PortWatcher."

        # For each vector log, save one figure
        for output_port_name, log_sink in self._drake_vector_logs.items():
            # Decide on whether or not there are "components"
            # in this output port (i.e., multiple values in a list output; different from a vector with multiple dimensions)
            port_component_name = None
            if output_port_name != self.port.get_name():
                port_component_name = output_port_name
            
            # Call the plotter to save the figure
            self.plotter.save_figures(log_sink, diagram_context, port_component_name=port_component_name)

    def save_raw_data(self, diagram_context: Context):
        """
        *Description*

        Saves the raw data to a file.

        *Arguments*
        
        diagram_context: Context
            The context of the diagram.
        """
        # Test to see the number of logs we have to save data for
        n_vector_logs = len(list(self._drake_vector_logs))

        # Iterate through all of the available VectorLogSink's
        for output_port_name, log_sink in self._drake_vector_logs.items():
            # Collect Drake Log of Data
            log = log_sink.FindLog(diagram_context)

            port = self.port
            system_containing_port: LeafSystem = port.get_system()

            # Write the data to file
            # - time data
            time_data_file_name = self.file_manager.time_data_file_path(
                system_name=system_containing_port.get_name(),
                port_name=port.get_name()
            )

            log_times = log.sample_times()
            os.makedirs(time_data_file_name.parent, exist_ok=True)
            np.save(time_data_file_name, log_times)

            # - data values
            raw_data_file: Path = None
            if n_vector_logs == 1:
                raw_data_file = self.file_manager.raw_data_file_path(
                    system_name=system_containing_port.get_name(),
                    port_name=port.get_name()
                )
            else:
                raw_data_file = self.file_manager.raw_data_file_path(
                    system_name=system_containing_port.get_name(),
                    port_name=port.get_name(),
                    port_component_name=output_port_name,
                )

                if raw_data_file.parent.exists() is False:
                    raw_data_file.parent.mkdir(parents=True, exist_ok=True)

            log_data = log.data()
            np.save(raw_data_file, log_data)

            # Announce the saving of the raw data
        



import networkx as nx
import numpy as np
from pydrake.systems.framework import LeafSystem, Diagram
from pydrake.systems.primitives import VectorLogSink
from typing import Tuple, Union

Performer = Union[LeafSystem, Diagram]

# A MotionPlan is either a networkx DiGraph or a numpy array
# The MotionPlan is either a graph that CONTAINS the plan (DiGraph) or
# a path in the configuration space (numpy array)
MotionPlan = Union[nx.DiGraph, np.ndarray]

MotionPlanningResult = Union[
    Tuple[MotionPlan, int],
    MotionPlan,
]

SupportedLogger = Union[VectorLogSink]




DEFAULT_BROM_DIR = "./brom" #: Default directory for BROM Drake data storage
DEFAULT_WATCHER_DIR = DEFAULT_BROM_DIR + "/watcher" #: Default subdirectory for :py:class:`PortWatcher<brom_drake.PortWatcher.port_watcher.PortWatcher>` data
DEFAULT_PLOT_DIR = DEFAULT_WATCHER_DIR + "/plots" #: Default subdirectory for :py:class:`PortWatcher<brom_drake.PortWatcher.port_watcher.PortWatcher>` plots
DEFAULT_RAW_DATA_DIR = DEFAULT_WATCHER_DIR + "/raw_data" #: Default subdirectory for :py:class:`PortWatcher<brom_drake.PortWatcher.port_watcher.PortWatcher>` raw data
DEFAULT_BROM_MODELS_DIR = DEFAULT_BROM_DIR + "/models" #: Default subdirectory for BROM Drake models
DEFAULT_NETWORKX_FSM_DIR = DEFAULT_BROM_DIR + "/networkx_fsm" #: Default subdirectory for NetworkX FSM data


from .add_watcher import (
    add_watcher,
    add_watcher_and_build,
    parse_list_of_simplified_targets
)

__all__ = [
    'add_watcher', 'add_watcher_and_build', 'parse_list_of_simplified_targets',
]